@@ -1,4 +1,4 @@
-/**
+/*
  * Licensed to the Apache Software Foundation (ASF) under one or more
  * contributor license agreements.  See the NOTICE file distributed with
  * this work for additional information regarding copyright ownership.
@@ -18,21 +18,36 @@ package org.apache.qpid.jms;
 
 import java.net.URI;
 import java.net.URISyntaxException;
+import java.security.AccessController;
+import java.security.PrivilegedAction;
 import java.util.Map;
 
 import javax.jms.Connection;
 import javax.jms.ConnectionFactory;
 import javax.jms.ExceptionListener;
-import javax.jms.IllegalStateException;
+import javax.jms.JMSContext;
 import javax.jms.JMSException;
+import javax.jms.JMSRuntimeException;
 import javax.jms.QueueConnection;
 import javax.jms.QueueConnectionFactory;
 import javax.jms.TopicConnection;
 import javax.jms.TopicConnectionFactory;
+import javax.net.ssl.SSLContext;
 
 import org.apache.qpid.jms.exceptions.JmsExceptionSupport;
 import org.apache.qpid.jms.jndi.JNDIStorable;
+import org.apache.qpid.jms.meta.JmsConnectionId;
 import org.apache.qpid.jms.meta.JmsConnectionInfo;
+import org.apache.qpid.jms.policy.JmsDefaultDeserializationPolicy;
+import org.apache.qpid.jms.policy.JmsDefaultMessageIDPolicy;
+import org.apache.qpid.jms.policy.JmsDefaultPrefetchPolicy;
+import org.apache.qpid.jms.policy.JmsDefaultPresettlePolicy;
+import org.apache.qpid.jms.policy.JmsDefaultRedeliveryPolicy;
+import org.apache.qpid.jms.policy.JmsDeserializationPolicy;
+import org.apache.qpid.jms.policy.JmsMessageIDPolicy;
+import org.apache.qpid.jms.policy.JmsPrefetchPolicy;
+import org.apache.qpid.jms.policy.JmsPresettlePolicy;
+import org.apache.qpid.jms.policy.JmsRedeliveryPolicy;
 import org.apache.qpid.jms.provider.Provider;
 import org.apache.qpid.jms.provider.ProviderFactory;
 import org.apache.qpid.jms.util.IdGenerator;
@@ -48,22 +63,32 @@ import org.slf4j.LoggerFactory;
 public class JmsConnectionFactory extends JNDIStorable implements ConnectionFactory, QueueConnectionFactory, TopicConnectionFactory {
 
     private static final Logger LOG = LoggerFactory.getLogger(JmsConnectionFactory.class);
+
     private static final String CLIENT_ID_PROP = "clientID";
+    private static final String DEFAULT_REMOTE_HOST = "localhost";
+    private static final String DEFAULT_REMOTE_PORT = "5672";
 
     public static final String REMOTE_URI_PROP = "remoteURI";
 
+    private static String DEFAULT_REMOTE_URI;
+
     private URI remoteURI;
     private String username;
     private String password;
     private String clientID;
     private boolean forceAsyncSend;
-    private boolean alwaysSyncSend;
-    private boolean sendAcksAsync;
+    private boolean forceSyncSend;
+    private boolean forceAsyncAcks;
     private boolean localMessagePriority;
     private boolean localMessageExpiry = true;
+    private boolean receiveLocalOnly;
+    private boolean receiveNoWaitLocalOnly;
+    private boolean populateJMSXUserID;
     private String queuePrefix = null;
     private String topicPrefix = null;
     private boolean validatePropertyNames = true;
+    private boolean awaitClientID = true;
+    private boolean useDaemonThread = false;
     private long sendTimeout = JmsConnectionInfo.DEFAULT_SEND_TIMEOUT;
     private long requestTimeout = JmsConnectionInfo.DEFAULT_REQUEST_TIMEOUT;
     private long closeTimeout = JmsConnectionInfo.DEFAULT_CLOSE_TIMEOUT;
@@ -74,8 +99,13 @@ public class JmsConnectionFactory extends JNDIStorable implements ConnectionFact
     private String connectionIDPrefix;
     private ExceptionListener exceptionListener;
 
-    private JmsPrefetchPolicy prefetchPolicy = new JmsPrefetchPolicy();
-    private JmsRedeliveryPolicy redeliveryPolicy = new JmsRedeliveryPolicy();
+    private JmsPrefetchPolicy prefetchPolicy = new JmsDefaultPrefetchPolicy();
+    private JmsRedeliveryPolicy redeliveryPolicy = new JmsDefaultRedeliveryPolicy();
+    private JmsPresettlePolicy presettlePolicy = new JmsDefaultPresettlePolicy();
+    private JmsMessageIDPolicy messageIDPolicy = new JmsDefaultMessageIDPolicy();
+    private JmsDeserializationPolicy deserializationPolicy = new JmsDefaultDeserializationPolicy();
+
+    private SSLContext sslContext;
 
     public JmsConnectionFactory() {
     }
@@ -86,17 +116,7 @@ public class JmsConnectionFactory extends JNDIStorable implements ConnectionFact
     }
 
     public JmsConnectionFactory(String remoteURI) {
-        this(createURI(remoteURI));
-    }
-
-    public JmsConnectionFactory(URI remoteURI) {
-        setRemoteURI(remoteURI.toString());
-    }
-
-    public JmsConnectionFactory(String userName, String password, URI remoteURI) {
-        setUsername(userName);
-        setPassword(password);
-        setRemoteURI(remoteURI.toString());
+        setRemoteURI(remoteURI);
     }
 
     public JmsConnectionFactory(String userName, String password, String remoteURI) {
@@ -105,6 +125,16 @@ public class JmsConnectionFactory extends JNDIStorable implements ConnectionFact
         setRemoteURI(remoteURI);
     }
 
+    public JmsConnectionFactory(URI remoteURI) {
+        setRemoteURI(remoteURI == null ? null : remoteURI.toString());
+    }
+
+    public JmsConnectionFactory(String userName, String password, URI remoteURI) {
+        setUsername(userName);
+        setPassword(password);
+        setRemoteURI(remoteURI == null ? null : remoteURI.toString());
+    }
+
     @Override
     protected Map<String, String> buildFromProperties(Map<String, String> props) {
         // Apply the remoteURI in a consistent order before
@@ -128,132 +158,175 @@ public class JmsConnectionFactory extends JNDIStorable implements ConnectionFact
         }
     }
 
-    /**
-     * @return a TopicConnection
-     * @throws JMSException
-     * @see javax.jms.TopicConnectionFactory#createTopicConnection()
-     */
     @Override
     public TopicConnection createTopicConnection() throws JMSException {
         return createTopicConnection(getUsername(), getPassword());
     }
 
-    /**
-     * @param username
-     * @param password
-     * @return a TopicConnection
-     * @throws JMSException
-     * @see javax.jms.TopicConnectionFactory#createTopicConnection(java.lang.String,
-     *      java.lang.String)
-     */
     @Override
     public TopicConnection createTopicConnection(String username, String password) throws JMSException {
+        JmsTopicConnection connection = null;
+
         try {
-            String connectionId = getConnectionIdGenerator().generateId();
+            JmsConnectionInfo connectionInfo = configureConnectionInfo(username, password);
             Provider provider = createProvider(remoteURI);
-            JmsTopicConnection result = new JmsTopicConnection(connectionId, provider, getClientIdGenerator());
-            return configureConnection(result, username, password);
+
+            connection = new JmsTopicConnection(connectionInfo, provider);
+            connection.setExceptionListener(exceptionListener);
+            connection.connect();
         } catch (Exception e) {
+            if (connection != null) {
+                try {
+                    connection.close();
+                } catch (Throwable ignored) {}
+            }
             throw JmsExceptionSupport.create(e);
         }
+
+        return connection;
     }
 
-    /**
-     * @return a Connection
-     * @throws JMSException
-     * @see javax.jms.ConnectionFactory#createConnection()
-     */
     @Override
     public Connection createConnection() throws JMSException {
         return createConnection(getUsername(), getPassword());
     }
 
-    /**
-     * @param username
-     * @param password
-     * @return Connection
-     * @throws JMSException
-     * @see javax.jms.ConnectionFactory#createConnection(java.lang.String, java.lang.String)
-     */
     @Override
     public Connection createConnection(String username, String password) throws JMSException {
+        JmsConnection connection = null;
+
         try {
-            String connectionId = getConnectionIdGenerator().generateId();
+            JmsConnectionInfo connectionInfo = configureConnectionInfo(username, password);
             Provider provider = createProvider(remoteURI);
-            JmsConnection result = new JmsConnection(connectionId, provider, getClientIdGenerator());
-            return configureConnection(result, username, password);
+
+            connection = new JmsConnection(connectionInfo, provider);
+            connection.setExceptionListener(exceptionListener);
+            connection.connect();
         } catch (Exception e) {
-            throw JmsExceptionSupport.create("Failed to create connection to: " + getRemoteURI(), e);
+            if (connection != null) {
+                try {
+                    connection.close();
+                } catch (Throwable ignored) {}
+            }
+            throw JmsExceptionSupport.create(e);
         }
+
+        return connection;
     }
 
-    /**
-     * @return a QueueConnection
-     * @throws JMSException
-     * @see javax.jms.QueueConnectionFactory#createQueueConnection()
-     */
     @Override
     public QueueConnection createQueueConnection() throws JMSException {
         return createQueueConnection(getUsername(), getPassword());
     }
 
-    /**
-     * @param username
-     * @param password
-     * @return a QueueConnection
-     * @throws JMSException
-     * @see javax.jms.QueueConnectionFactory#createQueueConnection(java.lang.String,
-     *      java.lang.String)
-     */
     @Override
     public QueueConnection createQueueConnection(String username, String password) throws JMSException {
+        JmsQueueConnection connection = null;
+
         try {
-            String connectionId = getConnectionIdGenerator().generateId();
+            JmsConnectionInfo connectionInfo = configureConnectionInfo(username, password);
             Provider provider = createProvider(remoteURI);
-            JmsQueueConnection result = new JmsQueueConnection(connectionId, provider, getClientIdGenerator());
-            return configureConnection(result, username, password);
+
+            connection = new JmsQueueConnection(connectionInfo, provider);
+            connection.setExceptionListener(exceptionListener);
+            connection.connect();
         } catch (Exception e) {
+            if (connection != null) {
+                try {
+                    connection.close();
+                } catch (Throwable ignored) {}
+            }
             throw JmsExceptionSupport.create(e);
         }
+
+        return connection;
     }
 
-    protected <T extends JmsConnection> T configureConnection(T connection, String username, String password) throws JMSException {
+    protected JmsConnectionInfo configureConnectionInfo(String username, String password) throws JMSException {
         try {
-
             Map<String, String> properties = PropertyUtil.getProperties(this);
-            // We must ensure that we apply the clientID last, since setting it on
-            // the Connection object provokes establishing the underlying connection.
-            boolean setClientID = false;
+            // Pull out the clientID prop, we need to act differently according to
+            // whether it was set in the URI or not.
+            boolean userSpecifiedClientId = false;
             if (properties.containsKey(CLIENT_ID_PROP)) {
-                setClientID = true;
+                userSpecifiedClientId = true;
                 properties.remove(CLIENT_ID_PROP);
             }
 
-            PropertyUtil.setProperties(connection, properties);
-            connection.setExceptionListener(exceptionListener);
-            connection.setUsername(username);
-            connection.setPassword(password);
-            connection.setConfiguredURI(remoteURI);
-            if (setClientID){
-                connection.setClientID(clientID);
+            String connectionId = getConnectionIdGenerator().generateId();
+
+            // Copy the configured policies before applying URI options that
+            // might make additional configuration changes.
+            JmsConnectionInfo connectionInfo = new JmsConnectionInfo(new JmsConnectionId(connectionId));
+
+            connectionInfo.setMessageIDPolicy(messageIDPolicy.copy());
+            connectionInfo.setPrefetchPolicy(prefetchPolicy.copy());
+            connectionInfo.setPresettlePolicy(presettlePolicy.copy());
+            connectionInfo.setRedeliveryPolicy(redeliveryPolicy.copy());
+            connectionInfo.setDeserializationPolicy(deserializationPolicy.copy());
+            connectionInfo.setSslContextOverride(sslContext);
+
+            // Set properties to make additional configuration changes
+            PropertyUtil.setProperties(connectionInfo, properties);
+
+            // Ensure we use the user and pass provided for this particular connection
+            connectionInfo.setUsername(username);
+            connectionInfo.setPassword(password);
+
+            connectionInfo.setConfiguredURI(remoteURI);
+
+            // Set the ClientID/container-id details
+            if (userSpecifiedClientId) {
+                connectionInfo.setClientId(clientID, true);
+            } else {
+                connectionInfo.setClientId(getClientIdGenerator().generateId(), false);
             }
 
-            return connection;
+            return connectionInfo;
         } catch (Exception e) {
             throw JmsExceptionSupport.create(e);
         }
     }
 
+    //----- JMSContext Creation methods --------------------------------------//
+
+    @Override
+    public JMSContext createContext() {
+        return createContext(getUsername(), getPassword(), JMSContext.AUTO_ACKNOWLEDGE);
+    }
+
+    @Override
+    public JMSContext createContext(int sessionMode) {
+        return createContext(getUsername(), getPassword(), sessionMode);
+    }
+
+    @Override
+    public JMSContext createContext(String username, String password) {
+        return createContext(username, password, JMSContext.AUTO_ACKNOWLEDGE);
+    }
+
+    @Override
+    public JMSContext createContext(String username, String password, int sessionMode) {
+        validateSessionMode(sessionMode);
+        try {
+            JmsConnection connection = (JmsConnection) createConnection(username, password);
+            return new JmsContext(connection, sessionMode);
+        } catch (JMSException jmse) {
+            throw JmsExceptionSupport.createRuntimeException(jmse);
+        }
+    }
+
+    //----- Internal Support Methods -----------------------------------------//
+
     protected Provider createProvider(URI remoteURI) throws Exception {
-        if(remoteURI == null) {
-            throw new IllegalStateException("No remoteURI has been provided");
+        if (remoteURI == null) {
+            remoteURI = new URI(getDefaultRemoteAddress());
         }
 
         Provider result = null;
 
         try {
             result = ProviderFactory.create(remoteURI);
-            result.connect();
         } catch (Exception ex) {
             LOG.error("Failed to create JMS Provider instance for: {}", remoteURI.getScheme());
             LOG.trace("Error: ", ex);
@@ -289,15 +362,13 @@ public class JmsConnectionFactory extends JNDIStorable implements ConnectionFact
         this.connectionIdGenerator = connectionIdGenerator;
     }
 
-    //////////////////////////////////////////////////////////////////////////
-    // Property getters and setters
-    //////////////////////////////////////////////////////////////////////////
+    //----- Property Access Methods ------------------------------------------//
 
     /**
      * @return the remoteURI
      */
     public String getRemoteURI() {
-        return this.remoteURI != null ? this.remoteURI.toString() : "";
+        return this.remoteURI != null ? this.remoteURI.toString() : getDefaultRemoteAddress();
     }
 
     /**
@@ -305,47 +376,55 @@ public class JmsConnectionFactory extends JNDIStorable implements ConnectionFact
      *        the remoteURI to set
      */
     public void setRemoteURI(String remoteURI) {
-        if (remoteURI == null) {
-            throw new IllegalArgumentException("remoteURI cannot be null");
+        if (remoteURI == null || remoteURI.isEmpty()) {
+            throw new IllegalArgumentException("Invalid URI: cannot be null or empty");
         }
         this.remoteURI = createURI(remoteURI);
 
-        try {
-            if (this.remoteURI.getQuery() != null) {
-                Map<String, String> map = PropertyUtil.parseQuery(this.remoteURI.getQuery());
-                Map<String, String> jmsOptionsMap = PropertyUtil.filterProperties(map, "jms.");
+        if (this.remoteURI.getRawUserInfo() != null) {
+            throw new IllegalArgumentException("The supplied URI cannot contain a User-Info section");
+        }
 
-                Map<String, String> unused = PropertyUtil.setProperties(this, jmsOptionsMap);
-                if (!unused.isEmpty()) {
-                    String msg = ""
-                        + " Not all jms options could be set on the ConnectionFactory."
-                        + " Check the options are spelled correctly."
-                        + " Unused parameters=[" + unused + "]."
-                        + " This connection factory cannot be started.";
-                    throw new IllegalArgumentException(msg);
-                } else {
-                    this.remoteURI = PropertyUtil.replaceQuery(this.remoteURI, map);
-                }
-            } else if (URISupport.isCompositeURI(this.remoteURI)) {
+        try {
+            if (URISupport.isCompositeURI(this.remoteURI)) {
                 CompositeData data = URISupport.parseComposite(this.remoteURI);
-                Map<String, String> jmsOptionsMap = PropertyUtil.filterProperties(data.getParameters(), "jms.");
-                Map<String, String> unused = PropertyUtil.setProperties(this, jmsOptionsMap);
-                if (!unused.isEmpty()) {
-                    String msg = ""
-                        + " Not all jms options could be set on the ConnectionFactory."
-                        + " Check the options are spelled correctly."
-                        + " Unused parameters=[" + unused + "]."
-                        + " This connection factory cannot be started.";
-                    throw new IllegalArgumentException(msg);
-                } else {
-                    this.remoteURI = data.toURI();
-                }
+                applyURIOptions(data.getParameters());
+                this.remoteURI = data.toURI();
+            } else if (this.remoteURI.getRawQuery() != null) {
+                Map<String, String> map = PropertyUtil.parseQuery(this.remoteURI);
+                applyURIOptions(map);
+                this.remoteURI = PropertyUtil.replaceQuery(this.remoteURI, map);
             }
         } catch (Exception e) {
             throw new IllegalArgumentException(e.getMessage());
         }
     }
 
+    private void applyURIOptions(Map<String, String> options) throws IllegalArgumentException {
+        Map<String, String> jmsOptionsMap = PropertyUtil.filterProperties(options, "jms.");
+        Map<String, String> unused = PropertyUtil.setProperties(this, jmsOptionsMap);
+        if (!unused.isEmpty()) {
+            String msg = ""
+                + " Not all jms options could be set on the ConnectionFactory."
+                + " Check the options are spelled correctly."
+                + " Unused parameters=[" + unused + "]."
+                + " This connection factory cannot be started.";
+            throw new IllegalArgumentException(msg);
+        }
+    }
+
+    private static void validateSessionMode(int mode) {
+        switch (mode) {
+            case JMSContext.AUTO_ACKNOWLEDGE:
+            case JMSContext.CLIENT_ACKNOWLEDGE:
+            case JMSContext.DUPS_OK_ACKNOWLEDGE:
+            case JMSContext.SESSION_TRANSACTED:
+                return;
+            default:
+                throw new JMSRuntimeException("Invalid Session Mode: " + mode);
+        }
+    }
+
     /**
      * @return the user name used for connection authentication.
      */
@@ -376,6 +455,27 @@ public class JmsConnectionFactory extends JNDIStorable implements ConnectionFact
         this.password = password;
     }
 
+    /**
+     * Returns true if the client should always send messages using a synchronous
+     * send operation regardless of persistence mode, or inside a transaction.
+     *
+     * @return true if sends should always be done synchronously.
+     */
+    public boolean isForceSyncSend() {
+        return forceSyncSend;
+    }
+
+    /**
+     * Configures whether or not the client will always send messages synchronously or not
+     * regardless of other factors that might result in an asynchronous send.
+     *
+     * @param forceSyncSend
+     *        if true sends are always done synchronously.
+     */
+    public void setForceSyncSend(boolean forceSyncSend) {
+        this.forceSyncSend = forceSyncSend;
+    }
+
     public boolean isForceAsyncSend() {
         return forceAsyncSend;
     }
@@ -503,6 +603,10 @@ public class JmsConnectionFactory extends JNDIStorable implements ConnectionFact
     }
 
     public void setPrefetchPolicy(JmsPrefetchPolicy prefetchPolicy) {
+        if (prefetchPolicy == null) {
+            prefetchPolicy = new JmsDefaultPrefetchPolicy();
+        }
+
         this.prefetchPolicy = prefetchPolicy;
     }
 
@@ -522,9 +626,74 @@ public class JmsConnectionFactory extends JNDIStorable implements ConnectionFact
      *        The new redeliveryPolicy to set
      */
     public void setRedeliveryPolicy(JmsRedeliveryPolicy redeliveryPolicy) {
+        if (redeliveryPolicy == null) {
+            redeliveryPolicy = new JmsDefaultRedeliveryPolicy();
+        }
         this.redeliveryPolicy = redeliveryPolicy;
     }
 
+    /**
+     * @return the presettlePolicy that is currently configured.
+     */
+    public JmsPresettlePolicy getPresettlePolicy() {
+        return presettlePolicy;
+    }
+
+    /**
+     * Sets the JmsPresettlePolicy that is applied to MessageProducers.
+     *
+     * @param presettlePolicy
+     *      the presettlePolicy to use by connections created from this factory.
+     */
+    public void setPresettlePolicy(JmsPresettlePolicy presettlePolicy) {
+        if (presettlePolicy == null) {
+            presettlePolicy = new JmsDefaultPresettlePolicy();
+        }
+        this.presettlePolicy = presettlePolicy;
+    }
+
+    /**
+     * @return the messageIDPolicy that is currently configured.
+     */
+    public JmsMessageIDPolicy getMessageIDPolicy() {
+        return messageIDPolicy;
+    }
+
+    /**
+     * Sets the JmsMessageIDPolicy that is use to configure the JmsMessageIDBuilder
+     * that is assigned to any new MessageProducer created from Connection instances
+     * that this factory has created.
+     *
+     * @param messageIDPolicy
+     *      the messageIDPolicy to use by connections created from this factory.
+     */
+    public void setMessageIDPolicy(JmsMessageIDPolicy messageIDPolicy) {
+        if (messageIDPolicy == null) {
+            messageIDPolicy = new JmsDefaultMessageIDPolicy();
+        }
+        this.messageIDPolicy = messageIDPolicy;
+    }
+
+    /**
+     * @return the deserializationPolicy that is currently configured.
+     */
+    public JmsDeserializationPolicy getDeserializationPolicy() {
+        return deserializationPolicy;
+    }
+
+    /**
+     * Sets the JmsDeserializationPolicy that is applied when a new connection is created.
+     *
+     * @param deserializationPolicy
+     *      the deserializationPolicy that will be applied to new connections.
+     */
+    public void setDeserializationPolicy(JmsDeserializationPolicy deserializationPolicy) {
+        if (deserializationPolicy == null) {
+            deserializationPolicy = new JmsDefaultDeserializationPolicy();
+        }
+        this.deserializationPolicy = deserializationPolicy;
+    }
+
     /**
      * @return the currently configured client ID prefix for auto-generated client IDs.
      */
@@ -570,6 +739,9 @@ public class JmsConnectionFactory extends JNDIStorable implements ConnectionFact
      * It is possible to set the clientID on the Connection itself immediately after
      * creation if no value has been set via the factory that created it, which will
      * allow the factory to create multiple open connections at a time.
+     *
+     * @param clientID
+     *      The clientID to assign when creating a new connection.
      */
     public void setClientID(String clientID) {
         this.clientID = clientID;
@@ -608,32 +780,11 @@ public class JmsConnectionFactory extends JNDIStorable implements ConnectionFact
         this.exceptionListener = exceptionListener;
     }
 
-    /**
-     * Returns true if the client should always send messages using a synchronous
-     * send operation regardless of persistence mode, or inside a transaction.
-     *
-     * @return true if sends should always be done synchronously.
-     */
-    public boolean isAlwaysSyncSend() {
-        return alwaysSyncSend;
-    }
-
-    /**
-     * Configures whether or not the client will always send messages synchronously or not
-     * regardless of other factors that might result in an asynchronous send.
-     *
-     * @param alwaysSyncSend
-     *        if true sends are always done synchronously.
-     */
-    public void setAlwaysSyncSend(boolean alwaysSyncSend) {
-        this.alwaysSyncSend = alwaysSyncSend;
-    }
-
     /**
      * @return true if consumer acknowledgments are sent asynchronously or not.
      */
-    public boolean isSendAcksAsync() {
-        return sendAcksAsync;
+    public boolean isForceAsyncAcks() {
+        return forceAsyncAcks;
     }
 
     /**
@@ -642,11 +793,11 @@ public class JmsConnectionFactory extends JNDIStorable implements ConnectionFact
      * performance of a consumer but opens up the possibility of a missed message
      * acknowledge should the connection be unstable.
      *
-     * @param sendAcksAsync
+     * @param forceAsyncAcks
      *        true to have the client send all message acknowledgments asynchronously.
      */
-    public void setSendAcksAsync(boolean sendAcksAsync) {
-        this.sendAcksAsync = sendAcksAsync;
+    public void setForceAsyncAcks(boolean forceAsyncAcks) {
+        this.forceAsyncAcks = forceAsyncAcks;
     }
 
     /**
@@ -667,4 +818,176 @@ public class JmsConnectionFactory extends JNDIStorable implements ConnectionFact
     public void setLocalMessageExpiry(boolean localMessageExpiry) {
         this.localMessageExpiry = localMessageExpiry;
     }
+
+    public boolean isReceiveLocalOnly() {
+        return receiveLocalOnly;
+    }
+
+    /**
+     * Controls whether the client only checks its local message buffer when using
+     * receive calls with a timeout, or will instead drain remaining credit from the
+     * remote peer to ensure there are really no messages available if the
+     * timeout expires before a message arrives in the consumers local buffer.
+     *
+     * @param receiveLocalOnly
+     *        true if receive calls with a timeout should only check the local message buffer.
+     */
+    public void setReceiveLocalOnly(boolean receiveLocalOnly) {
+        this.receiveLocalOnly = receiveLocalOnly;
+    }
+
+    public boolean isReceiveNoWaitLocalOnly() {
+        return receiveNoWaitLocalOnly;
+    }
+
+    /**
+     * Controls whether the client only checks its local message buffer when using
+     * receiveNoWait calls, or will instead drain remaining credit from the
+     * remote peer synchronously to ensure there are really no messages available
+     * that have yet to arrive in the consumers local buffer.
+     *
+     * @param receiveNoWaitLocalOnly
+     *        true if receiveNoWait calls should only check the local message buffer.
+     */
+    public void setReceiveNoWaitLocalOnly(boolean receiveNoWaitLocalOnly) {
+        this.receiveNoWaitLocalOnly = receiveNoWaitLocalOnly;
+    }
+
+    public boolean isPopulateJMSXUserID() {
+        return populateJMSXUserID;
+    }
+
+    /**
+     * Controls whether message sent from the Connection will have the JMSXUserID message
+     * property populated with the authenticated user ID of the Connection.  When false all
+     * messages sent from the Connection will not carry any value in the JMSXUserID property
+     * regardless of it being manually set on the Message to prevent a client spoofing the
+     * JMSXUserID value.
+     *
+     * @param populateJMSXUserID
+     *      true if message sent from this connection should have the JMSXUserID value populated.
+     */
+    public void setPopulateJMSXUserID(boolean populateJMSXUserID) {
+        this.populateJMSXUserID = populateJMSXUserID;
+    }
+
+    /**
+     * Sets an SSLContext to use when creating an SSL/TLS secured connection with this factory.
+     * The URI must still be configured to indicate a secure connection should be created.
+     * Using this method overrides the effect of URI/System property configuration relating
+     * to the location/credentials/type of SSL key/trust stores and whether to trust all
+     * certificates or use a particular keyAlias.
+     *
+     * @param sslContext
+     *      the sslContext to use, or null to respect the URI/System property configuration again.
+     */
+    public void setSslContext(SSLContext sslContext) {
+        this.sslContext = sslContext;
+    }
+
+    public boolean isAwaitClientID() {
+        return awaitClientID;
+    }
+
+    /**
+     * Controls whether the client will wait for a ClientID value to be set or the Connection
+     * to be used before it will attempt to complete the AMQP connection Open process.
+     * <p>
+     * By default a newly created Connection that does not have a ClientID configured in the URI
+     * will wait until a call to setClientID or some other interaction with the Connection API
+     * occurs before finishing the AMQP connection Open process with the remote peer. In some
+     * cases if this takes too long the remote can disconnect as a way of defending against
+     * denial of service attacks. If the user does not plan on setting a ClientID then this
+     * option allows for immediate AMQP connection Open completion and avoids the case where
+     * the remote peer might drop the Connection if it isn't used promptly.
+     * <p>
+     * This value defaults to true.
+     *
+     * @param awaitClientID
+     * 		the whether to wait for the client ID to be set before activating the connection.
+     */
+    public void setAwaitClientID(boolean awaitClientID) {
+        this.awaitClientID = awaitClientID;
+    }
+
+    public boolean isUseDaemonThread() {
+        return useDaemonThread;
+    }
+
+    /**
+     * Sets whether the Connection created will ensure that there is at least one non-daemon
+     * thread running at all times.
+     *
+     * @param useDaemonThread
+     * 		controls whether the Connection maintains a non-daemon thread.
+     */
+    public void setUseDaemonThread(boolean useDaemonThread) {
+        this.useDaemonThread = useDaemonThread;
+    }
+
+    //----- Static Methods ---------------------------------------------------//
+
+    /**
+     * @return the default remote address to connect to in the event that none was set.
+     */
+    public static String getDefaultRemoteAddress() {
+
+        if (DEFAULT_REMOTE_URI != null) {
+            return DEFAULT_REMOTE_URI;
+        }
+
+        synchronized (JmsConnectionFactory.class) {
+
+            if (DEFAULT_REMOTE_URI != null) {
+                return DEFAULT_REMOTE_URI;
+            }
+
+            String host = null;
+            String port = null;
+            String remoteAddress = null;
+
+            try {
+                host = AccessController.doPrivileged(new PrivilegedAction<String>() {
+                    @Override
+                    public String run() {
+                        String result = System.getProperty("org.apache.qpid.jms.REMOTE_HOST", DEFAULT_REMOTE_HOST);
+                        result = (result == null || result.isEmpty()) ? DEFAULT_REMOTE_HOST : result;
+                        return result;
+                    }
+                });
+                port = AccessController.doPrivileged(new PrivilegedAction<String>() {
+                    @Override
+                    public String run() {
+                        String result = System.getProperty("org.apache.qpid.jms.REMOTE_PORT", DEFAULT_REMOTE_PORT);
+                        result = (result == null || result.isEmpty()) ? DEFAULT_REMOTE_PORT : result;
+                        return result;
+                    }
+                });
+            } catch (Throwable e) {
+                LOG.debug("Failed to look up System properties for remote host and port", e);
+            }
+
+            host = (host == null || host.isEmpty()) ? DEFAULT_REMOTE_HOST : host;
+            port = (port == null || port.isEmpty()) ? DEFAULT_REMOTE_PORT : port;
+
+            final String defaultURL = "amqp://" + host + ":" + port;
+
+            try {
+                remoteAddress = AccessController.doPrivileged(new PrivilegedAction<String>() {
+                    @Override
+                    public String run() {
+                        String result = System.getProperty("org.apache.qpid.jms.REMOTE_URI", defaultURL);
+                        result = (result == null || result.isEmpty()) ? defaultURL : result;
+                        return result;
+                    }
+                });
+            } catch (Throwable e) {
+                LOG.debug("Failed to look up System property for remote URI", e);
+            }
+
+            DEFAULT_REMOTE_URI = (remoteAddress == null || remoteAddress.isEmpty()) ? defaultURL : remoteAddress;
+        }
+
+        return DEFAULT_REMOTE_URI;
+    }
 }
