@@ -1,4 +1,4 @@
-/**
+/*
  * Licensed to the Apache Software Foundation (ASF) under one or more
  * contributor license agreements.  See the NOTICE file distributed with
  * this work for additional information regarding copyright ownership.
@@ -16,9 +16,10 @@
  */
 package org.apache.qpid.jms;
 
-import java.util.List;
-import java.util.concurrent.Callable;
+import static org.apache.qpid.jms.message.JmsMessageSupport.lookupAckTypeForDisposition;
+
 import java.util.concurrent.atomic.AtomicBoolean;
+import java.util.concurrent.atomic.AtomicReference;
 import java.util.concurrent.locks.Lock;
 import java.util.concurrent.locks.ReentrantLock;
 
@@ -26,6 +27,7 @@ import javax.jms.IllegalStateException;
 import javax.jms.JMSException;
 import javax.jms.Message;
 import javax.jms.MessageConsumer;
+import javax.jms.MessageFormatException;
 import javax.jms.MessageListener;
 import javax.jms.Session;
 
@@ -34,6 +36,10 @@ import org.apache.qpid.jms.message.JmsInboundMessageDispatch;
 import org.apache.qpid.jms.message.JmsMessage;
 import org.apache.qpid.jms.meta.JmsConsumerId;
 import org.apache.qpid.jms.meta.JmsConsumerInfo;
+import org.apache.qpid.jms.meta.JmsResource.ResourceState;
+import org.apache.qpid.jms.policy.JmsDeserializationPolicy;
+import org.apache.qpid.jms.policy.JmsPrefetchPolicy;
+import org.apache.qpid.jms.policy.JmsRedeliveryPolicy;
 import org.apache.qpid.jms.provider.Provider;
 import org.apache.qpid.jms.provider.ProviderConstants.ACK_TYPE;
 import org.apache.qpid.jms.provider.ProviderFuture;
@@ -46,7 +52,7 @@ import org.slf4j.LoggerFactory;
 /**
  * implementation of a JMS Message Consumer
  */
-public class JmsMessageConsumer implements MessageConsumer, JmsMessageAvailableConsumer, JmsMessageDispatcher {
+public class JmsMessageConsumer implements AutoCloseable, MessageConsumer, JmsMessageAvailableConsumer, JmsMessageDispatcher {
 
     private static final Logger LOG = LoggerFactory.getLogger(JmsMessageConsumer.class);
 
@@ -55,48 +61,27 @@ public class JmsMessageConsumer implements MessageConsumer, JmsMessageAvailableC
     protected JmsConsumerInfo consumerInfo;
     protected final int acknowledgementMode;
     protected final AtomicBoolean closed = new AtomicBoolean();
-    protected boolean started;
-    protected MessageListener messageListener;
-    protected JmsMessageAvailableListener availableListener;
+    protected volatile MessageListener messageListener;
+    protected volatile JmsMessageAvailableListener availableListener;
     protected final MessageQueue messageQueue;
     protected final Lock lock = new ReentrantLock();
+    protected final Lock dispatchLock = new ReentrantLock();
     protected final AtomicBoolean suspendedConnection = new AtomicBoolean();
-    protected final AtomicBoolean delivered = new AtomicBoolean();
-    protected Exception failureCause;
+    protected final AtomicReference<Throwable> failureCause = new AtomicReference<>();
+    protected final MessageDeliverTask deliveryTask = new MessageDeliverTask();
 
-    /**
-     * Create a non-durable MessageConsumer
-     *
-     * @param consumerId
-     * @param session
-     * @param destination
-     * @param selector
-     * @param noLocal
-     * @throws JMSException
-     */
     protected JmsMessageConsumer(JmsConsumerId consumerId, JmsSession session, JmsDestination destination,
                                  String selector, boolean noLocal) throws JMSException {
         this(consumerId, session, destination, null, selector, noLocal);
     }
 
-    /**
-     * Create a MessageConsumer which could be durable.
-     *
-     * @param consumerId
-     * @param session
-     * @param destination
-     * @param name
-     * @param selector
-     * @param noLocal
-     * @throws JMSException
-     */
     protected JmsMessageConsumer(JmsConsumerId consumerId, JmsSession session, JmsDestination destination,
                                  String name, String selector, boolean noLocal) throws JMSException {
         this.session = session;
         this.connection = session.getConnection();
-        this.acknowledgementMode = session.acknowledgementMode();
+        this.acknowledgementMode = isBrowser() ? Session.AUTO_ACKNOWLEDGE : session.acknowledgementMode();
 
-        if(destination.isTemporary()) {
+        if (destination.isTemporary()) {
             connection.checkConsumeFromTemporaryDestination((JmsTemporaryDestination) destination);
         }
 
@@ -106,27 +91,40 @@ public class JmsMessageConsumer implements MessageConsumer, JmsMessageAvailableC
             this.messageQueue = new FifoMessageQueue();
         }
 
-        JmsPrefetchPolicy policy = connection.getPrefetchPolicy();
-        JmsRedeliveryPolicy redeliveryPolicy = connection.getRedeliveryPolicy().copy();
+        JmsPrefetchPolicy prefetchPolicy = session.getPrefetchPolicy();
+        JmsRedeliveryPolicy redeliveryPolicy = session.getRedeliveryPolicy().copy();
+        JmsDeserializationPolicy deserializationPolicy = session.getDeserializationPolicy().copy();
 
-        consumerInfo = new JmsConsumerInfo(consumerId);
-        consumerInfo.setClientId(connection.getClientID());
+        consumerInfo = new JmsConsumerInfo(consumerId, messageQueue);
+        consumerInfo.setExplicitClientID(connection.isExplicitClientID());
         consumerInfo.setSelector(selector);
+        consumerInfo.setDurable(isDurableSubscription());
         consumerInfo.setSubscriptionName(name);
+        consumerInfo.setShared(isSharedSubscription());
         consumerInfo.setDestination(destination);
         consumerInfo.setAcknowledgementMode(acknowledgementMode);
         consumerInfo.setNoLocal(noLocal);
         consumerInfo.setBrowser(isBrowser());
-        consumerInfo.setPrefetchSize(getConfiguredPrefetch(destination, policy));
+        consumerInfo.setPrefetchSize(
+            prefetchPolicy.getConfiguredPrefetch(session, destination, isDurableSubscription(), isBrowser()));
         consumerInfo.setRedeliveryPolicy(redeliveryPolicy);
         consumerInfo.setLocalMessageExpiry(connection.isLocalMessageExpiry());
+        consumerInfo.setPresettle(session.getPresettlePolicy().isConsumerPresttled(session, destination));
+        consumerInfo.setDeserializationPolicy(deserializationPolicy);
 
-        session.getConnection().createResource(consumerInfo);
+        session.add(this);
+        try {
+            session.getConnection().createResource(consumerInfo);
+        } catch (JMSException jmse) {
+            session.remove(this);
+            throw jmse;
+        }
     }
 
     public void init() throws JMSException {
-        session.add(this);
-        startConsumerResource();
+        if (!isPullConsumer()){
+            startConsumerResource();
+        }
     }
 
     private void startConsumerResource() throws JMSException {
@@ -138,35 +136,10 @@ public class JmsMessageConsumer implements MessageConsumer, JmsMessageAvailableC
         }
     }
 
-    /**
-     * @throws JMSException
-     * @see javax.jms.MessageConsumer#close()
-     */
     @Override
     public void close() throws JMSException {
         if (!closed.get()) {
-            session.getTransactionContext().addSynchronization(new JmsTxSynchronization() {
-
-                @Override
-                public boolean validate(JmsTransactionContext context) throws Exception {
-                    if (!context.isInTransaction() || !delivered.get()) {
-                        doClose();
-                        return false;
-                    }
-
-                    return true;
-                }
-
-                @Override
-                public void afterCommit() throws Exception {
-                    doClose();
-                }
-
-                @Override
-                public void afterRollback() throws Exception {
-                    doClose();
-                }
-            });
+            doClose();
         }
     }
 
@@ -174,7 +147,7 @@ public class JmsMessageConsumer implements MessageConsumer, JmsMessageAvailableC
      * Called to initiate shutdown of Producer resources and request that the remote
      * peer remove the registered producer.
      *
-     * @throws JMSException
+     * @throws JMSException if an error occurs during the consumer close operation.
      */
     protected void doClose() throws JMSException {
         shutdown();
@@ -186,36 +159,26 @@ public class JmsMessageConsumer implements MessageConsumer, JmsMessageAvailableC
      * to be sent to the remote peer.  This is most commonly needed when the parent
      * Session is closing.
      *
-     * @throws JMSException
+     * @throws JMSException if an error occurs during shutdown.
      */
     protected void shutdown() throws JMSException {
         shutdown(null);
     }
 
-    protected void shutdown(Exception cause) throws JMSException {
+    protected void shutdown(Throwable cause) throws JMSException {
         if (closed.compareAndSet(false, true)) {
-            failureCause = cause;
+            consumerInfo.setState(ResourceState.CLOSED);
+            setFailureCause(cause);
             session.remove(this);
             stop(true);
         }
     }
 
-    /**
-     * @return a Message or null if closed during the operation
-     * @throws JMSException
-     * @see javax.jms.MessageConsumer#receive()
-     */
     @Override
     public Message receive() throws JMSException {
         return receive(0);
     }
 
-    /**
-     * @param timeout
-     * @return a Message or null
-     * @throws JMSException
-     * @see javax.jms.MessageConsumer#receive(long)
-     */
     @Override
     public Message receive(long timeout) throws JMSException {
         checkClosed();
@@ -226,19 +189,67 @@ public class JmsMessageConsumer implements MessageConsumer, JmsMessageAvailableC
             timeout = -1;
         }
 
-        return copy(ackFromReceive(dequeue(timeout)));
+        return copy(ackFromReceive(dequeue(timeout, connection.isReceiveLocalOnly())));
     }
 
-    /**
-     * @return a Message or null
-     * @throws JMSException
-     * @see javax.jms.MessageConsumer#receiveNoWait()
-     */
     @Override
     public Message receiveNoWait() throws JMSException {
         checkClosed();
         checkMessageListener();
-        return copy(ackFromReceive(dequeue(0)));
+
+        return copy(ackFromReceive(dequeue(0, connection.isReceiveNoWaitLocalOnly())));
+    }
+
+    /**
+     * Reads the next available message for this consumer and returns the body of that message
+     * if the type requested matches that of the message.  The amount of time this method blocks
+     * is based on the timeout value.
+     *
+     *   {@literal timeout < 0} then it blocks until a message is received.
+     *   {@literal timeout = 0} then it returns the body immediately or null if none available.
+     *   {@literal timeout > 0} then it blocks up to timeout amount of time.
+     *
+     * @param desired
+     *      The type to assign the body of the message to for return.
+     * @param timeout
+     *      The time to wait for an incoming message before this method returns null.
+     *
+     * @return the assigned body of the next available message or null if the consumer is closed
+     *         or the specified timeout elapses.
+     *
+     * @throws MessageFormatException if the message body cannot be assigned to the requested type.
+     * @throws JMSException if an error occurs while receiving the next message.
+     */
+    public <T> T receiveBody(Class<T> desired, long timeout) throws JMSException {
+        checkClosed();
+        checkMessageListener();
+
+        T messageBody = null;
+        JmsInboundMessageDispatch envelope = null;
+
+        try {
+            envelope = dequeue(timeout, connection.isReceiveLocalOnly());
+            if (envelope != null) {
+                messageBody = envelope.getMessage().getBody(desired);
+            }
+        } catch (MessageFormatException mfe) {
+            // Should behave as if receiveBody never happened in these modes.
+            if (acknowledgementMode == Session.AUTO_ACKNOWLEDGE ||
+                acknowledgementMode == Session.DUPS_OK_ACKNOWLEDGE) {
+
+                envelope.setEnqueueFirst(true);
+                onInboundMessage(envelope);
+                envelope = null;
+            }
+
+            throw mfe;
+        } finally {
+            if (envelope != null) {
+                ackFromReceive(envelope);
+            }
+        }
+
+        return messageBody;
     }
 
     /**
@@ -252,10 +263,16 @@ public class JmsMessageConsumer implements MessageConsumer, JmsMessageAvailableC
      * This method may consume messages that are expired or exceed a configured
      * delivery count value but will continue to wait for the configured timeout.
      *
-     * @throws JMSException
-     * @return null if we timeout or if the consumer is closed.
+     * @param localCheckOnly
+     *          if false, try pulling a message if a >= 0 timeout expires with no message arriving
+     *
+     * @return null if we timeout or if the consumer is closed concurrently.
+     *
+     * @throws JMSException if an error occurs during the dequeue.
      */
-    private JmsInboundMessageDispatch dequeue(long timeout) throws JMSException {
+    private JmsInboundMessageDispatch dequeue(long timeout, boolean localCheckOnly) throws JMSException {
+        boolean pullConsumer = isPullConsumer();
+        boolean pullForced = pullConsumer;
 
         try {
             long deadline = 0;
@@ -263,47 +280,56 @@ public class JmsMessageConsumer implements MessageConsumer, JmsMessageAvailableC
                 deadline = System.currentTimeMillis() + timeout;
             }
 
-            sendPullCommand(timeout);
+            performPullIfRequired(timeout, false);
 
             while (true) {
                 JmsInboundMessageDispatch envelope = null;
-                if (isPullConsumer()) {
-                    envelope = messageQueue.dequeue(-1);
+                if (pullForced || pullConsumer) {
+                    // Any waiting was done by the pull request, try immediate retrieval from the queue.
+                    envelope = messageQueue.dequeue(0);
                 } else {
                     envelope = messageQueue.dequeue(timeout);
                 }
 
+                if (getFailureCause() != null) {
+                    LOG.debug("{} receive failed: {}", getConsumerId(), getFailureCause().getMessage());
+                    throw JmsExceptionSupport.create(getFailureCause());
+                }
+
                 if (envelope == null) {
-
-                    if (failureCause != null && !messageQueue.isClosed() && timeout == 0) {
-                        LOG.debug("{} receive failed: {}", getConsumerId(), failureCause.getMessage());
-                        throw JmsExceptionSupport.create(failureCause);
-                    }
-
-                    if (timeout == 0 || messageQueue.isClosed()) {
+                    if ((timeout == 0 && (pullForced || localCheckOnly)) || pullConsumer || messageQueue.isClosed()) {
                         return null;
                     } else if (timeout > 0) {
                         timeout = Math.max(deadline - System.currentTimeMillis(), 0);
                     }
 
-                    sendPullCommand(timeout);
-                } else if (envelope.getMessage() == null) {
-                    LOG.trace("{} no message was available for this consumer: {}", getConsumerId());
-                    return null;
+                    if (timeout >= 0 && !localCheckOnly) {
+                        // We don't do this for receive with no timeout since it
+                        // is redundant: zero-prefetch consumers already pull, and
+                        // the rest block indefinitely on the local messageQueue.
+                        pullForced = true;
+                        if (performPullIfRequired(timeout, true)) {
+                            startConsumerResource();
+                            // We refresh credit if it is a prefetching consumer, since the
+                            // pull drained it. Processing acks can open the credit window, but
+                            // not in all cases, and if we didn't get a message it would stay
+                            // closed until future pulls were performed.
+                        }
+                    }
                 } else if (consumeExpiredMessage(envelope)) {
                     LOG.trace("{} filtered expired message: {}", getConsumerId(), envelope);
                     doAckExpired(envelope);
                     if (timeout > 0) {
                         timeout = Math.max(deadline - System.currentTimeMillis(), 0);
                     }
-                    sendPullCommand(timeout);
+                    performPullIfRequired(timeout, false);
                 } else if (redeliveryExceeded(envelope)) {
                     LOG.debug("{} filtered message with excessive redelivery count: {}", getConsumerId(), envelope);
-                    doAckUndeliverable(envelope);
+                    applyRedeliveryPolicyOutcome(envelope);
                     if (timeout > 0) {
                         timeout = Math.max(deadline - System.currentTimeMillis(), 0);
                     }
-                    sendPullCommand(timeout);
+                    performPullIfRequired(timeout, false);
                 } else {
                     if (LOG.isTraceEnabled()) {
                         LOG.trace(getConsumerId() + " received message: " + envelope);
@@ -330,25 +356,37 @@ public class JmsMessageConsumer implements MessageConsumer, JmsMessageAvailableC
 
         JmsRedeliveryPolicy redeliveryPolicy = consumerInfo.getRedeliveryPolicy();
         return redeliveryPolicy != null &&
-               redeliveryPolicy.getMaxRedeliveries() != JmsRedeliveryPolicy.DEFAULT_MAX_REDELIVERIES &&
-               redeliveryPolicy.getMaxRedeliveries() < envelope.getRedeliveryCount();
+               redeliveryPolicy.getMaxRedeliveries(getDestination()) >= 0 &&
+               redeliveryPolicy.getMaxRedeliveries(getDestination()) < envelope.getRedeliveryCount();
     }
 
     protected void checkClosed() throws IllegalStateException {
         if (closed.get()) {
             IllegalStateException jmsEx = null;
 
-            if (failureCause == null) {
+            if (getFailureCause() == null) {
                 jmsEx = new IllegalStateException("The MessageConsumer is closed");
             } else {
                 jmsEx = new IllegalStateException("The MessageConsumer was closed due to an unrecoverable error.");
-                jmsEx.initCause(failureCause);
+                jmsEx.initCause(getFailureCause());
             }
 
             throw jmsEx;
         }
     }
 
+    void setFailureCause(Throwable failureCause) {
+        this.failureCause.set(failureCause);
+    }
+
+    Throwable getFailureCause() {
+        if (failureCause.get() == null) {
+            return session.getFailureCause();
+        }
+
+        return failureCause.get();
+    }
+
     JmsMessage copy(final JmsInboundMessageDispatch envelope) throws JMSException {
         if (envelope == null || envelope.getMessage() == null) {
             return null;
@@ -366,17 +404,13 @@ public class JmsMessageConsumer implements MessageConsumer, JmsMessageAvailableC
             } else {
                 doAckConsumed(envelope);
             }
-
-            // Tags that we have delivered and can't close if in a TX Session.
-            delivered.set(true);
         }
         return envelope;
     }
 
     private JmsInboundMessageDispatch doAckConsumed(final JmsInboundMessageDispatch envelope) throws JMSException {
-        checkClosed();
         try {
-            session.acknowledge(envelope, ACK_TYPE.CONSUMED);
+            session.acknowledge(envelope, ACK_TYPE.ACCEPTED);
         } catch (JMSException ex) {
             session.onException(ex);
             throw ex;
@@ -396,16 +430,17 @@ public class JmsMessageConsumer implements MessageConsumer, JmsMessageAvailableC
 
     private void doAckExpired(final JmsInboundMessageDispatch envelope) throws JMSException {
         try {
-            session.acknowledge(envelope, ACK_TYPE.EXPIRED);
+            session.acknowledge(envelope, ACK_TYPE.MODIFIED_FAILED_UNDELIVERABLE);
         } catch (JMSException ex) {
             session.onException(ex);
             throw ex;
         }
     }
 
-    private void doAckUndeliverable(final JmsInboundMessageDispatch envelope) throws JMSException {
+    private void applyRedeliveryPolicyOutcome(final JmsInboundMessageDispatch envelope) throws JMSException {
         try {
-            session.acknowledge(envelope, ACK_TYPE.POISONED);
+            JmsRedeliveryPolicy redeliveryPolicy = consumerInfo.getRedeliveryPolicy();
+            session.acknowledge(envelope, lookupAckTypeForDisposition(redeliveryPolicy.getOutcome(getDestination())));
         } catch (JMSException ex) {
             session.onException(ex);
             throw ex;
@@ -433,17 +468,7 @@ public class JmsMessageConsumer implements MessageConsumer, JmsMessageAvailableC
         lock.lock();
         try {
             if (acknowledgementMode == Session.CLIENT_ACKNOWLEDGE) {
-                envelope.getMessage().setAcknowledgeCallback(new Callable<Void>() {
-                    @Override
-                    public Void call() throws Exception {
-                        if (session.isClosed()) {
-                            throw new javax.jms.IllegalStateException("Session closed.");
-                        }
-                        session.acknowledge();
-                        envelope.getMessage().setAcknowledgeCallback(null);
-                        return null;
-                    }
-                });
+                envelope.getMessage().setAcknowledgeCallback(new JmsAcknowledgeCallback(session));
             }
 
             if (envelope.isEnqueueFirst()) {
@@ -452,14 +477,14 @@ public class JmsMessageConsumer implements MessageConsumer, JmsMessageAvailableC
                 this.messageQueue.enqueue(envelope);
             }
 
-            if (this.messageListener != null && this.started) {
-                session.getExecutor().execute(new MessageDeliverTask());
-            } else {
-                if (availableListener != null) {
-                    session.getExecutor().execute(new Runnable() {
+            if (session.isStarted() && messageQueue.isRunning()) {
+                if (messageListener != null) {
+                    session.getDispatcherExecutor().execute(deliveryTask);
+                } else if (availableListener != null) {
+                    session.getDispatcherExecutor().execute(new Runnable() {
                         @Override
                         public void run() {
-                            if (session.isStarted()) {
+                            if (messageQueue.isRunning()) {
                                 availableListener.onMessageAvailable(JmsMessageConsumer.this);
                             }
                         }
@@ -474,9 +499,10 @@ public class JmsMessageConsumer implements MessageConsumer, JmsMessageAvailableC
     public void start() {
         lock.lock();
         try {
-            this.started = true;
-            this.messageQueue.start();
-            drainMessageQueueToListener();
+            if (!messageQueue.isRunning()) {
+                this.messageQueue.start();
+                drainMessageQueueToListener();
+            }
         } finally {
             lock.unlock();
         }
@@ -487,9 +513,9 @@ public class JmsMessageConsumer implements MessageConsumer, JmsMessageAvailableC
     }
 
     private void stop(boolean closeMessageQueue) {
+        dispatchLock.lock();
         lock.lock();
         try {
-            this.started = false;
             if (closeMessageQueue) {
                 this.messageQueue.close();
             } else {
@@ -497,40 +523,32 @@ public class JmsMessageConsumer implements MessageConsumer, JmsMessageAvailableC
             }
         } finally {
             lock.unlock();
+            dispatchLock.unlock();
         }
     }
 
     void suspendForRollback() throws JMSException {
         stop();
 
-        session.getConnection().stopResource(consumerInfo);
+        try {
+            session.getConnection().stopResource(consumerInfo);
+        } finally {
+            if (session.getTransactionContext().isActiveInThisContext(getConsumerId())) {
+                messageQueue.clear();
+            }
+        }
     }
 
     void resumeAfterRollback() throws JMSException {
-        if (!this.messageQueue.isEmpty()) {
-            List<JmsInboundMessageDispatch> drain = this.messageQueue.removeAll();
-            for (JmsInboundMessageDispatch envelope : drain) {
-                doAckReleased(envelope);
-            }
-            drain.clear();
-        }
-
         start();
-
         startConsumerResource();
     }
 
-    void drainMessageQueueToListener() {
-        if (this.messageListener != null && this.started) {
-            session.getExecutor().execute(new MessageDeliverTask());
-        }
-    }
-
     /**
      * @return the id
      */
     public JmsConsumerId getConsumerId() {
-        return this.consumerInfo.getConsumerId();
+        return this.consumerInfo.getId();
     }
 
     /**
@@ -546,27 +564,26 @@ public class JmsMessageConsumer implements MessageConsumer, JmsMessageAvailableC
         return this.messageListener;
     }
 
-    /**
-     * @param listener
-     * @throws JMSException
-     * @see javax.jms.MessageConsumer#setMessageListener(javax.jms.MessageListener)
-     */
     @Override
     public void setMessageListener(MessageListener listener) throws JMSException {
         checkClosed();
-        if (consumerInfo.getPrefetchSize() == 0) {
-            throw new JMSException("Illegal prefetch size of zero. This setting is not supported" +
-                                   "for asynchronous consumers please set a value of at least 1");
+
+        dispatchLock.lock();
+        try {
+            messageListener = listener;
+            consumerInfo.setListener(listener != null);
+
+            if (listener != null) {
+                if (isPullConsumer()){
+                    startConsumerResource();
+                }
+                drainMessageQueueToListener();
+            }
+        } finally {
+            dispatchLock.unlock();
         }
-        this.messageListener = listener;
-        drainMessageQueueToListener();
     }
 
-    /**
-     * @return the Message Selector
-     * @throws JMSException
-     * @see javax.jms.MessageConsumer#getMessageSelector()
-     */
     @Override
     public String getMessageSelector() throws JMSException {
         checkClosed();
@@ -605,6 +622,10 @@ public class JmsMessageConsumer implements MessageConsumer, JmsMessageAvailableC
         return false;
     }
 
+    public boolean isSharedSubscription() {
+        return false;
+    }
+
     public boolean isBrowser() {
         return false;
     }
@@ -628,25 +649,18 @@ public class JmsMessageConsumer implements MessageConsumer, JmsMessageAvailableC
     }
 
     protected void onConnectionRecovery(Provider provider) throws Exception {
-        ProviderFuture request = new ProviderFuture();
-        provider.create(consumerInfo, request);
-        request.sync();
+        if (consumerInfo.isOpen()) {
+            ProviderFuture request = new ProviderFuture();
+            provider.create(consumerInfo, request);
+            request.sync();
+        }
     }
 
     protected void onConnectionRecovered(Provider provider) throws Exception {
-        ProviderFuture request = new ProviderFuture();
-        provider.start(consumerInfo, request);
-        request.sync();
-
-        // For a pull consumer we want to stop the Queue waking an blocked receives and
-        // preventing a new one from blocking until it has reissued a new pull.
-        if (isPullConsumer()) {
-            lock.lock();
-            try {
-                messageQueue.stop();
-            } finally {
-                lock.unlock();
-            }
+        if (consumerInfo.isOpen()) {
+            ProviderFuture request = new ProviderFuture();
+            provider.start(consumerInfo, request);
+            request.sync();
         }
     }
 
@@ -654,7 +668,9 @@ public class JmsMessageConsumer implements MessageConsumer, JmsMessageAvailableC
     }
 
     /**
-     * Triggers a pull request from the connected Provider with the given timeout value.
+     * Triggers a pull request from the connected Provider with the given timeout value
+     * if the consumer is a pull consumer or requested to be treated as one, and the
+     * local queue is still running, and is currently empty.
      * <p>
      * The timeout value can be one of:
      * <br>
@@ -664,81 +680,114 @@ public class JmsMessageConsumer implements MessageConsumer, JmsMessageAvailableC
      *
      * @param timeout
      *        The amount of time the pull request should remain valid.
+     * @param treatAsPullConsumer
+     *        Treat the consumer as if it were a pull consumer, even if it isn't.
+     * @return true if a pull was performed, false if it was not.
      */
-    protected void sendPullCommand(long timeout) throws JMSException {
-        if (!messageQueue.isClosed() && messageQueue.isEmpty() && (getPrefetchSize() == 0 || isBrowser())) {
+    protected boolean performPullIfRequired(long timeout, boolean treatAsPullConsumer) throws JMSException {
+        if ((isPullConsumer() || treatAsPullConsumer) && messageQueue.isRunning() && messageQueue.isEmpty()) {
             connection.pull(getConsumerId(), timeout);
+            return true;
+        }
 
-            // Once a new pull has gone out check to see if the queue was stopped due to failover
-            // and restart it so that the receive calls can once again start blocking on dequeue.
-            lock.lock();
-            try {
-                if (started) {
-                    messageQueue.start();
-                }
-            } finally {
-                lock.unlock();
-            }
+        return false;
+    }
+
+    private void drainMessageQueueToListener() {
+        if (messageListener != null && session.isStarted() && messageQueue.isRunning()) {
+            session.getDispatcherExecutor().execute(new BoundedMessageDeliverTask(messageQueue.size()));
         }
     }
 
-    private int getConfiguredPrefetch(JmsDestination destination, JmsPrefetchPolicy policy) {
-        int prefetch = 0;
-        if (destination.isTopic()) {
-            if (isDurableSubscription()) {
-                prefetch = policy.getDurableTopicPrefetch();
-            } else {
-                prefetch = policy.getTopicPrefetch();
-            }
-        } else {
-            if (isBrowser()) {
-                prefetch = policy.getQueueBrowserPrefetch();
-            } else {
-                prefetch = policy.getQueuePrefetch();
+    private boolean deliverNextPending() {
+        if (session.isStarted() && messageQueue.isRunning() && messageListener != null) {
+            dispatchLock.lock();
+            try {
+                JmsInboundMessageDispatch envelope = messageQueue.dequeueNoWait();
+                if (envelope == null) {
+                    return false;
+                }
+
+                JmsMessage copy = null;
+
+                if (consumeExpiredMessage(envelope)) {
+                    LOG.trace("{} filtered expired message: {}", getConsumerId(), envelope);
+                    doAckExpired(envelope);
+                } else if (redeliveryExceeded(envelope)) {
+                    LOG.trace("{} filtered message with excessive redelivery count: {}", getConsumerId(), envelope);
+                    applyRedeliveryPolicyOutcome(envelope);
+                } else {
+                    boolean deliveryFailed = false;
+                    boolean autoAckOrDupsOk = acknowledgementMode == Session.AUTO_ACKNOWLEDGE ||
+                                              acknowledgementMode == Session.DUPS_OK_ACKNOWLEDGE;
+                    if (autoAckOrDupsOk) {
+                        copy = copy(doAckDelivered(envelope));
+                    } else {
+                        copy = copy(ackFromReceive(envelope));
+                    }
+                    session.clearSessionRecovered();
+
+                    try {
+                        messageListener.onMessage(copy);
+                    } catch (RuntimeException rte) {
+                        deliveryFailed = true;
+                    }
+
+                    if (autoAckOrDupsOk && !session.isSessionRecovered()) {
+                        if (!deliveryFailed) {
+                            doAckConsumed(envelope);
+                        } else {
+                            doAckReleased(envelope);
+                        }
+                    }
+                }
+            } catch (Exception e) {
+                // TODO - There are two cases where we can get an error here, one being
+                //        and error returned from the attempted ACK that was sent and the
+                //        other being an error while attempting to copy the incoming message.
+                //        We need to decide how to respond to these.
+                session.getConnection().onException(e);
+            } finally {
+                dispatchLock.unlock();
+
+                if (isPullConsumer()) {
+                    try {
+                        startConsumerResource();
+                    } catch (JMSException e) {
+                        LOG.error("Exception during credit replenishment for consumer listener {}", getConsumerId(), e);
+                    }
+                }
             }
         }
 
-        return prefetch;
+        return !messageQueue.isEmpty();
+    }
+
+    private final class BoundedMessageDeliverTask implements Runnable {
+
+        private final int deliveryCount;
+
+        public BoundedMessageDeliverTask(int deliveryCount) {
+            this.deliveryCount = deliveryCount;
+        }
+
+        @Override
+        public void run() {
+            int current = 0;
+
+            while (session.isStarted() && messageQueue.isRunning() && current++ < deliveryCount) {
+                if (!deliverNextPending()) {
+                    return;  // Another task already drained the queue.
+                }
+            }
+        }
     }
 
     private final class MessageDeliverTask implements Runnable {
+
         @Override
         public void run() {
-            JmsInboundMessageDispatch envelope;
-            while (session.isStarted() && (envelope = messageQueue.dequeueNoWait()) != null) {
-                try {
-                    JmsMessage copy = null;
-
-                    if (consumeExpiredMessage(envelope)) {
-                        LOG.trace("{} filtered expired message: {}", getConsumerId(), envelope);
-                        doAckExpired(envelope);
-                    } else if (redeliveryExceeded(envelope)) {
-                        LOG.trace("{} filtered message with excessive redelivery count: {}", getConsumerId(), envelope);
-                        doAckUndeliverable(envelope);
-                    } else {
-                        boolean autoAckOrDupsOk = acknowledgementMode == Session.AUTO_ACKNOWLEDGE ||
-                                                  acknowledgementMode == Session.DUPS_OK_ACKNOWLEDGE;
-                        if (autoAckOrDupsOk) {
-                            copy = copy(doAckDelivered(envelope));
-                        } else {
-                            copy = copy(ackFromReceive(envelope));
-                        }
-                        session.clearSessionRecovered();
-
-                        messageListener.onMessage(copy);
-
-                        if (autoAckOrDupsOk && !session.isSessionRecovered()) {
-                            doAckConsumed(envelope);
-                        }
-                    }
-                } catch (Exception e) {
-                    // TODO - We need to handle exception of on message with some other
-                    //        ack such as rejected and consider adding a redlivery policy
-                    //        to control when we might just poison the message with an ack
-                    //        of modified set to not deliverable here.
-                    session.getConnection().onException(e);
-                }
-            }
+            deliverNextPending();
         }
     }
 }
