@@ -1,4 +1,4 @@
-/**
+/*
  * Licensed to the Apache Software Foundation (ASF) under one or more
  * contributor license agreements.  See the NOTICE file distributed with
  * this work for additional information regarding copyright ownership.
@@ -18,7 +18,9 @@ package org.apache.qpid.jms;
 
 import java.util.concurrent.atomic.AtomicBoolean;
 import java.util.concurrent.atomic.AtomicLong;
+import java.util.concurrent.atomic.AtomicReference;
 
+import javax.jms.CompletionListener;
 import javax.jms.DeliveryMode;
 import javax.jms.Destination;
 import javax.jms.IllegalStateException;
@@ -27,20 +29,23 @@ import javax.jms.JMSException;
 import javax.jms.Message;
 import javax.jms.MessageProducer;
 
+import org.apache.qpid.jms.message.JmsMessageIDBuilder;
 import org.apache.qpid.jms.meta.JmsProducerId;
 import org.apache.qpid.jms.meta.JmsProducerInfo;
+import org.apache.qpid.jms.meta.JmsResource.ResourceState;
 import org.apache.qpid.jms.provider.Provider;
 import org.apache.qpid.jms.provider.ProviderFuture;
 
 /**
- * Implementation of a Jms MessageProducer
+ * Implementation of a JMS MessageProducer
  */
-public class JmsMessageProducer implements MessageProducer {
+public class JmsMessageProducer implements AutoCloseable, MessageProducer {
 
     protected final JmsSession session;
     protected final JmsConnection connection;
     protected JmsProducerInfo producerInfo;
-    protected final boolean flexibleDestination;
+    protected final boolean anonymousProducer;
+    protected long deliveryDelay = Message.DEFAULT_DELIVERY_DELAY;
     protected int deliveryMode = DeliveryMode.PERSISTENT;
     protected int priority = Message.DEFAULT_PRIORITY;
     protected long timeToLive = Message.DEFAULT_TIME_TO_LIVE;
@@ -48,25 +53,29 @@ public class JmsMessageProducer implements MessageProducer {
     protected boolean disableMessageId;
     protected boolean disableTimestamp;
     protected final AtomicLong messageSequence = new AtomicLong();
-    protected Exception failureCause;
+    protected final AtomicReference<Throwable> failureCause = new AtomicReference<>();
 
     protected JmsMessageProducer(JmsProducerId producerId, JmsSession session, JmsDestination destination) throws JMSException {
         this.session = session;
         this.connection = session.getConnection();
-        this.flexibleDestination = destination == null;
-        this.producerInfo = new JmsProducerInfo(producerId);
-        this.producerInfo.setDestination(destination);
+        this.anonymousProducer = destination == null;
 
-        session.getConnection().createResource(producerInfo);
+        JmsMessageIDBuilder messageIDBuilder =
+            session.getMessageIDPolicy().getMessageIDBuilder(session, destination);
+
+        this.producerInfo = new JmsProducerInfo(producerId, messageIDBuilder);
+        this.producerInfo.setDestination(destination);
+        this.producerInfo.setPresettle(session.getPresettlePolicy().isProducerPresttled(session, destination));
+
+        session.add(this);
+        try {
+            session.getConnection().createResource(producerInfo);
+        } catch (JMSException jmse) {
+            session.remove(this);
+            throw jmse;
+        }
     }
 
-    /**
-     * Close the producer
-     *
-     * @throws JMSException
-     *
-     * @see javax.jms.MessageProducer#close()
-     */
     @Override
     public void close() throws JMSException {
         if (!closed.get()) {
@@ -78,11 +87,12 @@ public class JmsMessageProducer implements MessageProducer {
      * Called to initiate shutdown of Producer resources and request that the remote
      * peer remove the registered producer.
      *
-     * @throws JMSException
+     * @throws JMSException if an internal error occurs during the close operation.
      */
     protected void doClose() throws JMSException {
+        session.checkIsCompletionThread();
         shutdown();
-        this.connection.destroyResource(producerInfo);
+        connection.destroyResource(producerInfo);
     }
 
     /**
@@ -90,204 +100,190 @@ public class JmsMessageProducer implements MessageProducer {
      * to be sent to the remote peer.  This is most commonly needed when the parent
      * Session is closing.
      *
-     * @throws JMSException
+     * @throws JMSException if an internal error occurs during the shutdown operation.
      */
     protected void shutdown() throws JMSException {
         shutdown(null);
     }
 
-    protected void shutdown(Exception cause) throws JMSException {
+    protected void shutdown(Throwable cause) throws JMSException {
         if (closed.compareAndSet(false, true)) {
-            failureCause = cause;
+            producerInfo.setState(ResourceState.CLOSED);
+            failureCause.set(cause);
             session.remove(this);
         }
     }
 
-    /**
-     * @return the delivery mode
-     * @throws JMSException
-     * @see javax.jms.MessageProducer#getDeliveryMode()
-     */
+    @Override
+    public long getDeliveryDelay() throws JMSException {
+        checkClosed();
+        return deliveryDelay;
+    }
+
     @Override
     public int getDeliveryMode() throws JMSException {
         checkClosed();
-        return this.deliveryMode;
+        return deliveryMode;
     }
 
-    /**
-     * @return the destination
-     * @throws JMSException
-     * @see javax.jms.MessageProducer#getDestination()
-     */
     @Override
     public Destination getDestination() throws JMSException {
         checkClosed();
-        return this.producerInfo.getDestination();
+        return producerInfo.getDestination();
     }
 
-    /**
-     * @return true if disableIds is set
-     * @throws JMSException
-     * @see javax.jms.MessageProducer#getDisableMessageID()
-     */
     @Override
     public boolean getDisableMessageID() throws JMSException {
         checkClosed();
-        return this.disableMessageId;
+        return disableMessageId;
     }
 
-    /**
-     * @return true if disable timestamp is set
-     * @throws JMSException
-     * @see javax.jms.MessageProducer#getDisableMessageTimestamp()
-     */
     @Override
     public boolean getDisableMessageTimestamp() throws JMSException {
         checkClosed();
-        return this.disableTimestamp;
+        return disableTimestamp;
     }
 
-    /**
-     * @return the priority
-     * @throws JMSException
-     * @see javax.jms.MessageProducer#getPriority()
-     */
     @Override
     public int getPriority() throws JMSException {
         checkClosed();
-        return this.priority;
+        return priority;
     }
 
-    /**
-     * @return timeToLive
-     * @throws JMSException
-     * @see javax.jms.MessageProducer#getTimeToLive()
-     */
     @Override
     public long getTimeToLive() throws JMSException {
         checkClosed();
-        return this.timeToLive;
+        return timeToLive;
     }
 
-    /**
-     * @param message
-     * @throws JMSException
-     * @see javax.jms.MessageProducer#send(javax.jms.Message)
-     */
     @Override
     public void send(Message message) throws JMSException {
-        send(message, this.deliveryMode, this.priority, this.timeToLive);
+        send(message, deliveryMode, priority, timeToLive);
     }
 
-    /**
-     * @param message
-     * @param deliveryMode
-     * @param priority
-     * @param timeToLive
-     * @throws JMSException
-     * @see javax.jms.MessageProducer#send(javax.jms.Message, int, int, long)
-     */
     @Override
     public void send(Message message, int deliveryMode, int priority, long timeToLive) throws JMSException {
         checkClosed();
 
-        if (flexibleDestination) {
+        if (anonymousProducer) {
             throw new UnsupportedOperationException("Using this method is not supported on producers created without an explicit Destination");
         }
 
-        sendMessage(producerInfo.getDestination(), message, deliveryMode, priority, timeToLive);
+        sendMessage(producerInfo.getDestination(), message, deliveryMode, priority, timeToLive, null);
     }
 
-    /**
-     * @param destination
-     * @param message
-     * @throws JMSException
-     * @see javax.jms.MessageProducer#send(javax.jms.Destination,
-     *      javax.jms.Message)
-     */
     @Override
     public void send(Destination destination, Message message) throws JMSException {
-        send(destination, message, this.deliveryMode, this.priority, this.timeToLive);
+        send(destination, message, deliveryMode, priority, timeToLive);
     }
 
-    /**
-     * @param destination
-     * @param message
-     * @param deliveryMode
-     * @param priority
-     * @param timeToLive
-     * @throws JMSException
-     * @see javax.jms.MessageProducer#send(javax.jms.Destination,
-     *      javax.jms.Message, int, int, long)
-     */
     @Override
     public void send(Destination destination, Message message, int deliveryMode, int priority, long timeToLive) throws JMSException {
         checkClosed();
 
-        if (!flexibleDestination) {
+        checkDestinationNotInvalid(destination);
+
+        if (!anonymousProducer) {
             throw new UnsupportedOperationException("Using this method is not supported on producers created with an explicit Destination.");
         }
 
-        sendMessage(destination, message, deliveryMode, priority, timeToLive);
+        sendMessage(destination, message, deliveryMode, priority, timeToLive, null);
     }
 
-    private void sendMessage(Destination destination, Message message, int deliveryMode, int priority, long timeToLive) throws JMSException {
-        if (destination == null) {
-            throw new InvalidDestinationException("Don't understand null destinations");
+    @Override
+    public void send(Message message, CompletionListener listener) throws JMSException {
+        send(message, deliveryMode, priority, timeToLive, listener);
+    }
+
+    @Override
+    public void send(Message message, int deliveryMode, int priority, long timeToLive, CompletionListener listener) throws JMSException {
+        checkClosed();
+
+        if (anonymousProducer) {
+            throw new UnsupportedOperationException("Using this method is not supported on producers created without an explicit Destination");
         }
 
-        this.session.send(this, destination, message, deliveryMode, priority, timeToLive, disableMessageId, disableTimestamp);
+        if (listener == null) {
+            throw new IllegalArgumentException("JmsCompletetionListener cannot be null");
+        }
+
+        sendMessage(producerInfo.getDestination(), message, deliveryMode, priority, timeToLive, listener);
+    }
+
+    @Override
+    public void send(Destination destination, Message message, CompletionListener listener) throws JMSException {
+        send(destination, message, this.deliveryMode, this.priority, this.timeToLive, listener);
+    }
+
+    @Override
+    public void send(Destination destination, Message message, int deliveryMode, int priority, long timeToLive, CompletionListener listener) throws JMSException {
+        checkClosed();
+
+        checkDestinationNotInvalid(destination);
+
+        if (!anonymousProducer) {
+            throw new UnsupportedOperationException("Using this method is not supported on producers created with an explicit Destination.");
+        }
+
+        if (listener == null) {
+            throw new IllegalArgumentException("JmsCompletetionListener cannot be null");
+        }
+
+        sendMessage(destination, message, deliveryMode, priority, timeToLive, listener);
+    }
+
+    private void checkDestinationNotInvalid(Destination destination) throws InvalidDestinationException {
+        if (destination == null) {
+            throw new InvalidDestinationException("Destination must not be null");
+        }
+    }
+
+    private void sendMessage(Destination destination, Message message, int deliveryMode, int priority, long timeToLive, CompletionListener listener) throws JMSException {
+        this.session.send(this, destination, message, deliveryMode, priority, timeToLive, disableMessageId, disableTimestamp, deliveryDelay, listener);
+    }
+
+    @Override
+    public void setDeliveryDelay(long deliveryDelay) throws JMSException {
+        checkClosed();
+        this.deliveryDelay = deliveryDelay;
     }
 
-    /**
-     * @param deliveryMode
-     * @throws JMSException
-     * @see javax.jms.MessageProducer#setDeliveryMode(int)
-     */
     @Override
     public void setDeliveryMode(int deliveryMode) throws JMSException {
         checkClosed();
-        this.deliveryMode = deliveryMode;
+        switch (deliveryMode) {
+            case DeliveryMode.PERSISTENT:
+            case DeliveryMode.NON_PERSISTENT:
+                this.deliveryMode = deliveryMode;
+                break;
+            default:
+                throw new JMSException(String.format("Invalid DeliveryMode specified: %d", deliveryMode));
+        }
     }
 
-    /**
-     * @param value
-     * @throws JMSException
-     * @see javax.jms.MessageProducer#setDisableMessageID(boolean)
-     */
     @Override
     public void setDisableMessageID(boolean value) throws JMSException {
         checkClosed();
         this.disableMessageId = value;
     }
 
-    /**
-     * @param value
-     * @throws JMSException
-     * @see javax.jms.MessageProducer#setDisableMessageTimestamp(boolean)
-     */
     @Override
     public void setDisableMessageTimestamp(boolean value) throws JMSException {
         checkClosed();
         this.disableTimestamp = value;
     }
 
-    /**
-     * @param defaultPriority
-     * @throws JMSException
-     * @see javax.jms.MessageProducer#setPriority(int)
-     */
     @Override
     public void setPriority(int defaultPriority) throws JMSException {
         checkClosed();
+
+        if (defaultPriority < 0 || defaultPriority > 9) {
+            throw new JMSException(String.format("Priority value given {%d} is out of range (0..9)", defaultPriority));
+        }
+
         this.priority = defaultPriority;
     }
 
-    /**
-     * @param timeToLive
-     * @throws JMSException
-     * @see javax.jms.MessageProducer#setTimeToLive(long)
-     */
     @Override
     public void setTimeToLive(long timeToLive) throws JMSException {
         checkClosed();
@@ -298,31 +294,55 @@ public class JmsMessageProducer implements MessageProducer {
      * @return the producer's assigned JmsProducerId.
      */
     protected JmsProducerId getProducerId() {
-        return this.producerInfo.getProducerId();
+        return this.producerInfo.getId();
     }
 
     /**
      * @return the next logical sequence for a Message sent from this Producer.
      */
     protected long getNextMessageSequence() {
-        return this.messageSequence.incrementAndGet();
+        return messageSequence.incrementAndGet();
     }
 
     protected void checkClosed() throws IllegalStateException {
         if (closed.get()) {
             IllegalStateException jmsEx = null;
 
-            if (failureCause == null) {
+            if (getFailureCause() == null) {
                 jmsEx = new IllegalStateException("The MessageProducer is closed");
             } else {
                 jmsEx = new IllegalStateException("The MessageProducer was closed due to an unrecoverable error.");
-                jmsEx.initCause(failureCause);
+                jmsEx.initCause(getFailureCause());
             }
 
             throw jmsEx;
         }
     }
 
+    protected boolean isPresettled() {
+        return producerInfo.isPresettle();
+    }
+
+    protected boolean isAnonymous() {
+        return anonymousProducer;
+    }
+
+    protected JmsMessageIDBuilder getMessageIDBuilder() {
+        return producerInfo.getMessageIDBuilder();
+    }
+
+    void setFailureCause(Throwable failureCause) {
+        this.failureCause.set(failureCause);
+    }
+
+    Throwable getFailureCause() {
+        if (failureCause.get() == null) {
+            return session.getFailureCause();
+        }
+
+        return failureCause.get();
+    }
+
     ////////////////////////////////////////////////////////////////////////////
     // Connection interruption handlers.
     ////////////////////////////////////////////////////////////////////////////
@@ -331,9 +351,11 @@ public class JmsMessageProducer implements MessageProducer {
     }
 
     protected void onConnectionRecovery(Provider provider) throws Exception {
-        ProviderFuture request = new ProviderFuture();
-        provider.create(producerInfo, request);
-        request.sync();
+        if (producerInfo.isOpen()) {
+            ProviderFuture request = new ProviderFuture();
+            provider.create(producerInfo, request);
+            request.sync();
+        }
     }
 
     protected void onConnectionRecovered(Provider provider) throws Exception {
