@@ -1,4 +1,4 @@
-/**
+/*
  * Licensed to the Apache Software Foundation (ASF) under one or more
  * contributor license agreements.  See the NOTICE file distributed with
  * this work for additional information regarding copyright ownership.
@@ -24,6 +24,7 @@ import javax.jms.JMSSecurityException;
 import javax.security.sasl.SaslException;
 
 import org.apache.qpid.jms.meta.JmsConnectionInfo;
+import org.apache.qpid.jms.provider.AsyncResult;
 import org.apache.qpid.jms.sasl.Mechanism;
 import org.apache.qpid.jms.sasl.SaslMechanismFinder;
 import org.apache.qpid.proton.engine.Sasl;
@@ -36,12 +37,16 @@ public class AmqpSaslAuthenticator {
     private final Sasl sasl;
     private final JmsConnectionInfo info;
     private Mechanism mechanism;
-    private Principal localPrincipal;
+    private final Principal localPrincipal;
     private Set<String> mechanismsRestriction;
+    private final AsyncResult authenticationRequest;
+    private boolean complete;
 
     /**
      * Create the authenticator and initialize it.
      *
+     * @param request
+     * 	      The initial request that is awaiting the result of the authentication process.
      * @param sasl
      *        The Proton SASL entry point this class will use to manage the authentication.
      * @param info
@@ -52,20 +57,22 @@ public class AmqpSaslAuthenticator {
      *        The possible mechanism(s) to which the client should restrict its
      *        mechanism selection to if offered by the server
      */
-    public AmqpSaslAuthenticator(Sasl sasl, JmsConnectionInfo info, Principal localPrincipal, String[] mechanismsRestriction) {
+    public AmqpSaslAuthenticator(AsyncResult request, Sasl sasl, JmsConnectionInfo info, Principal localPrincipal, String[] mechanismsRestriction) {
         this.sasl = sasl;
         this.info = info;
         this.localPrincipal = localPrincipal;
-        if(mechanismsRestriction != null) {
+        this.authenticationRequest = request;
+
+        if (mechanismsRestriction != null) {
             Set<String> mechs = new HashSet<String>();
-            for(int i = 0; i < mechanismsRestriction.length; i++) {
+            for (int i = 0; i < mechanismsRestriction.length; i++) {
                 String mech = mechanismsRestriction[i];
-                if(!mech.trim().isEmpty()) {
+                if (!mech.trim().isEmpty()) {
                     mechs.add(mech);
                 }
             }
 
-            if(!mechs.isEmpty()) {
+            if (!mechs.isEmpty()) {
                 this.mechanismsRestriction = mechs;
             }
         }
@@ -77,25 +84,53 @@ public class AmqpSaslAuthenticator {
      * successful authentication or a JMSSecurityException is thrown indicating that the
      * handshake failed.
      *
-     * @throws JMSSecurityException
+     * @return true if the authentication process completed.
      */
-    public boolean authenticate() throws JMSSecurityException {
+    public boolean authenticate() {
+        try {
+            switch (sasl.getState()) {
+                case PN_SASL_IDLE:
+                    handleSaslInit();
+                    break;
+                case PN_SASL_STEP:
+                    handleSaslStep();
+                    break;
+                case PN_SASL_FAIL:
+                    handleSaslFail();
+                    break;
+                case PN_SASL_PASS:
+                    authenticationRequest.onSuccess();
+                default:
+                    break;
+            }
+        } catch (JMSSecurityException result) {
+            authenticationRequest.onFailure(result);
+        }
+
+        complete = authenticationRequest.isComplete();
+
+        return complete;
+    }
+
+    public boolean isComplete() {
+       return complete;
+    }
+
+    public boolean wasSuccessful() throws IllegalStateException {
         switch (sasl.getState()) {
+            case PN_SASL_CONF:
             case PN_SASL_IDLE:
-                handleSaslInit();
-                break;
             case PN_SASL_STEP:
-                handleSaslStep();
                 break;
             case PN_SASL_FAIL:
-                handleSaslFail();
-                break;
+                return false;
             case PN_SASL_PASS:
                 return true;
             default:
+                break;
         }
 
-        return false;
+        throw new IllegalStateException("Authentication has not completed yet.");
     }
 
     private void handleSaslInit() throws JMSSecurityException {
@@ -114,13 +149,11 @@ public class AmqpSaslAuthenticator {
                         sasl.send(response, 0, response.length);
                     }
                 } else {
-                    // TODO - Better error message.
                     throw new JMSSecurityException("Could not find a suitable SASL mechanism for the remote peer using the available credentials.");
                 }
             }
         } catch (SaslException se) {
-            // TODO - Better error message.
-            JMSSecurityException jmsse = new JMSSecurityException("Exception while processing SASL init.");
+            JMSSecurityException jmsse = new JMSSecurityException("Exception while processing SASL init: " + se.getMessage());
             jmsse.setLinkedException(se);
             jmsse.initCause(se);
             throw jmsse;
@@ -136,8 +169,7 @@ public class AmqpSaslAuthenticator {
                 sasl.send(response, 0, response.length);
             }
         } catch (SaslException se) {
-            // TODO - Better error message.
-            JMSSecurityException jmsse = new JMSSecurityException("Exception while processing SASL step.");
+            JMSSecurityException jmsse = new JMSSecurityException("Exception while processing SASL step: " + se.getMessage());
             jmsse.setLinkedException(se);
             jmsse.initCause(se);
             throw jmsse;
@@ -145,7 +177,10 @@ public class AmqpSaslAuthenticator {
     }
 
     private void handleSaslFail() throws JMSSecurityException {
-        // TODO - Better error message.
-        throw new JMSSecurityException("Client failed to authenticate");
+        if (mechanism != null) {
+            throw new JMSSecurityException("Client failed to authenticate using SASL: " + mechanism.getName());
+        } else {
+            throw new JMSSecurityException("Client failed to authenticate");
+        }
     }
 }
