@@ -1,4 +1,4 @@
-/**
+/*
  * Licensed to the Apache Software Foundation (ASF) under one or more
  * contributor license agreements.  See the NOTICE file distributed with
  * this work for additional information regarding copyright ownership.
@@ -18,14 +18,14 @@ package org.apache.qpid.jms.provider.failover;
 
 import java.io.IOException;
 import java.net.URI;
-import java.net.URISyntaxException;
 import java.util.ArrayList;
+import java.util.Arrays;
 import java.util.LinkedHashMap;
 import java.util.List;
+import java.util.Locale;
 import java.util.Map;
-import java.util.concurrent.ExecutorService;
-import java.util.concurrent.Executors;
-import java.util.concurrent.ScheduledExecutorService;
+import java.util.concurrent.ScheduledFuture;
+import java.util.concurrent.ScheduledThreadPoolExecutor;
 import java.util.concurrent.ThreadFactory;
 import java.util.concurrent.TimeUnit;
 import java.util.concurrent.atomic.AtomicBoolean;
@@ -33,7 +33,11 @@ import java.util.concurrent.atomic.AtomicLong;
 import java.util.concurrent.atomic.AtomicReference;
 
 import javax.jms.JMSException;
+import javax.jms.JMSSecurityException;
+import javax.jms.TransactionRolledBackException;
 
+import org.apache.qpid.jms.JmsOperationTimedOutException;
+import org.apache.qpid.jms.JmsSendTimedOutException;
 import org.apache.qpid.jms.message.JmsInboundMessageDispatch;
 import org.apache.qpid.jms.message.JmsMessageFactory;
 import org.apache.qpid.jms.message.JmsOutboundMessageDispatch;
@@ -41,6 +45,7 @@ import org.apache.qpid.jms.meta.JmsConnectionInfo;
 import org.apache.qpid.jms.meta.JmsConsumerId;
 import org.apache.qpid.jms.meta.JmsResource;
 import org.apache.qpid.jms.meta.JmsSessionId;
+import org.apache.qpid.jms.meta.JmsTransactionInfo;
 import org.apache.qpid.jms.provider.AsyncResult;
 import org.apache.qpid.jms.provider.DefaultProviderListener;
 import org.apache.qpid.jms.provider.Provider;
@@ -49,6 +54,7 @@ import org.apache.qpid.jms.provider.ProviderFactory;
 import org.apache.qpid.jms.provider.ProviderFuture;
 import org.apache.qpid.jms.provider.ProviderListener;
 import org.apache.qpid.jms.provider.ProviderRedirectedException;
+import org.apache.qpid.jms.provider.WrappedAsyncResult;
 import org.apache.qpid.jms.util.IOExceptionSupport;
 import org.apache.qpid.jms.util.ThreadPoolUtils;
 import org.slf4j.Logger;
@@ -66,8 +72,12 @@ public class FailoverProvider extends DefaultProviderListener implements Provide
 
     public static final int UNLIMITED = -1;
 
+    private static final int UNDEFINED = -1;
+    private static final int DISABLED = 0;
+    private static final int MINIMUM_TIMEOUT = 1000;
+
     public static final int DEFAULT_MAX_RECONNECT_ATTEMPTS = UNLIMITED;
-    public static final int DEFAULT_STARTUP_MAX_RECONNECT_ATTEMPTS = UNLIMITED;
+    public static final int DEFAULT_STARTUP_MAX_RECONNECT_ATTEMPTS = UNDEFINED;
     public static final long DEFAULT_INITIAL_RECONNECT_DELAY = 0;
     public static final long DEFAULT_RECONNECT_DELAY = 10;
     public static final long DEFAULT_MAX_RECONNECT_DELAY = TimeUnit.SECONDS.toMillis(30);
@@ -78,9 +88,10 @@ public class FailoverProvider extends DefaultProviderListener implements Provide
     private ProviderListener listener;
     private Provider provider;
     private final FailoverUriPool uris;
+    private ScheduledFuture<?> requestTimeoutTask;
 
-    private final ExecutorService serializer;
-    private final ScheduledExecutorService connectionHub;
+    private final ScheduledThreadPoolExecutor serializer;
+    private final ScheduledThreadPoolExecutor connectionHub;
     private final AtomicBoolean closed = new AtomicBoolean();
     private final AtomicBoolean failed = new AtomicBoolean();
     private final AtomicBoolean closingConnection = new AtomicBoolean(false);
@@ -90,16 +101,17 @@ public class FailoverProvider extends DefaultProviderListener implements Provide
     private final AtomicReference<JmsMessageFactory> messageFactory = new AtomicReference<JmsMessageFactory>();
 
     // Current state of connection / reconnection
+    private boolean firstAttempt = true;
     private boolean firstConnection = true;
     private long reconnectAttempts;
     private long nextReconnectDelay = -1;
     private IOException failureCause;
     private URI connectedURI;
+    private volatile JmsConnectionInfo connectionInfo;
 
     // Timeout values configured via JmsConnectionInfo
-    private long connectTimeout = JmsConnectionInfo.DEFAULT_CONNECT_TIMEOUT;
     private long closeTimeout = JmsConnectionInfo.DEFAULT_CLOSE_TIMEOUT;
-    private long sendTimeout =  JmsConnectionInfo.DEFAULT_SEND_TIMEOUT;
+    private long sendTimeout = JmsConnectionInfo.DEFAULT_SEND_TIMEOUT;
     private long requestTimeout = JmsConnectionInfo.DEFAULT_REQUEST_TIMEOUT;
 
     // Configuration values.
@@ -112,6 +124,8 @@ public class FailoverProvider extends DefaultProviderListener implements Provide
     private int startupMaxReconnectAttempts = DEFAULT_STARTUP_MAX_RECONNECT_ATTEMPTS;
     private int warnAfterReconnectAttempts = DEFAULT_WARN_AFTER_RECONNECT_ATTEMPTS;
 
+    private FailoverServerListAction amqpOpenServerListAction = FailoverServerListAction.REPLACE;
+
     public FailoverProvider(Map<String, String> nestedOptions) {
         this(null, nestedOptions);
     }
@@ -123,7 +137,7 @@ public class FailoverProvider extends DefaultProviderListener implements Provide
     public FailoverProvider(List<URI> uris, Map<String, String> nestedOptions) {
         this.uris = new FailoverUriPool(uris, nestedOptions);
 
-        this.serializer = Executors.newSingleThreadScheduledExecutor(new ThreadFactory() {
+        serializer = new ScheduledThreadPoolExecutor(1, new ThreadFactory() {
 
             @Override
             public Thread newThread(Runnable runner) {
@@ -134,10 +148,13 @@ public class FailoverProvider extends DefaultProviderListener implements Provide
             }
         });
 
+        serializer.setExecuteExistingDelayedTasksAfterShutdownPolicy(false);
+        serializer.setContinueExistingPeriodicTasksAfterShutdownPolicy(false);
+
         // All Connection attempts happen in this schedulers thread.  Once a connection
         // is established it will hand the open connection back to the serializer thread
         // for state recovery.
-        this.connectionHub = Executors.newScheduledThreadPool(1, new ThreadFactory() {
+        connectionHub = new ScheduledThreadPoolExecutor(1, new ThreadFactory() {
 
             @Override
             public Thread newThread(Runnable runner) {
@@ -147,11 +164,15 @@ public class FailoverProvider extends DefaultProviderListener implements Provide
                 return serial;
             }
         });
+
+        connectionHub.setExecuteExistingDelayedTasksAfterShutdownPolicy(false);
+        connectionHub.setContinueExistingPeriodicTasksAfterShutdownPolicy(false);
     }
 
     @Override
-    public void connect() throws IOException {
+    public void connect(JmsConnectionInfo connectionInfo) throws IOException {
         checkClosed();
+        this.connectionInfo = connectionInfo;
         LOG.debug("Initiating initial connection attempt task");
         triggerReconnectionAttempt();
     }
@@ -180,6 +201,10 @@ public class FailoverProvider extends DefaultProviderListener implements Provide
                             request.onFailure(error);
                         }
 
+                        if (requestTimeoutTask != null) {
+                            requestTimeoutTask.cancel(false);
+                        }
+
                         if (provider != null) {
                             provider.close();
                         }
@@ -199,7 +224,7 @@ public class FailoverProvider extends DefaultProviderListener implements Provide
                 if (this.closeTimeout < 0) {
                     request.sync();
                 } else {
-                    request.sync(closeTimeout, TimeUnit.MILLISECONDS);
+                    request.sync(Math.max(MINIMUM_TIMEOUT, closeTimeout), TimeUnit.MILLISECONDS);
                 }
             } catch (IOException e) {
                 LOG.warn("Error caught while closing Provider: ", e.getMessage());
@@ -216,7 +241,8 @@ public class FailoverProvider extends DefaultProviderListener implements Provide
                 @Override
                 public void doTask() throws Exception {
                     JmsConnectionInfo connectionInfo = (JmsConnectionInfo) resource;
-                    connectTimeout = connectionInfo.getConnectTimeout();
+
+                    // Collect the timeouts we will handle in this provider.
                     closeTimeout = connectionInfo.getCloseTimeout();
                     sendTimeout = connectionInfo.getSendTimeout();
                     requestTimeout = connectionInfo.getRequestTimeout();
@@ -230,12 +256,25 @@ public class FailoverProvider extends DefaultProviderListener implements Provide
                 }
             };
         } else {
-            pending = new FailoverRequest(request) {
+            pending = new FailoverRequest(request, requestTimeout) {
                 @Override
                 public void doTask() throws Exception {
                     provider.create(resource, this);
                 }
 
+                @Override
+                public boolean succeedsWhenOffline() {
+                    if (resource instanceof JmsTransactionInfo) {
+                        // Tag as in-doubt and let recovery on reconnect sort it out.
+                        JmsTransactionInfo transactionInfo = (JmsTransactionInfo) resource;
+                        transactionInfo.setInDoubt(true);
+
+                        return true;
+                    } else {
+                        return false;
+                    }
+                }
+
                 @Override
                 public String toString() {
                     return "create -> " + resource;
@@ -249,7 +288,7 @@ public class FailoverProvider extends DefaultProviderListener implements Provide
     @Override
     public void start(final JmsResource resource, final AsyncResult request) throws IOException, JMSException {
         checkClosed();
-        final FailoverRequest pending = new FailoverRequest(request) {
+        final FailoverRequest pending = new FailoverRequest(request, requestTimeout) {
             @Override
             public void doTask() throws Exception {
                 provider.start(resource, this);
@@ -267,7 +306,7 @@ public class FailoverProvider extends DefaultProviderListener implements Provide
     @Override
     public void stop(final JmsResource resource, final AsyncResult request) throws IOException, JMSException {
         checkClosed();
-        final FailoverRequest pending = new FailoverRequest(request) {
+        final FailoverRequest pending = new FailoverRequest(request, requestTimeout) {
             @Override
             public void doTask() throws Exception {
                 provider.stop(resource, this);
@@ -285,7 +324,7 @@ public class FailoverProvider extends DefaultProviderListener implements Provide
     @Override
     public void destroy(final JmsResource resourceId, AsyncResult request) throws IOException, JMSException, UnsupportedOperationException {
         checkClosed();
-        final FailoverRequest pending = new FailoverRequest(request) {
+        final FailoverRequest pending = new FailoverRequest(request, requestTimeout) {
             @Override
             public void doTask() throws IOException, JMSException, UnsupportedOperationException {
                 if (resourceId instanceof JmsConnectionInfo) {
@@ -312,7 +351,7 @@ public class FailoverProvider extends DefaultProviderListener implements Provide
     @Override
     public void send(final JmsOutboundMessageDispatch envelope, AsyncResult request) throws IOException, JMSException {
         checkClosed();
-        final FailoverRequest pending = new FailoverRequest(request) {
+        final FailoverRequest pending = new FailoverRequest(request, sendTimeout) {
             @Override
             public void doTask() throws Exception {
                 provider.send(envelope, this);
@@ -322,18 +361,23 @@ public class FailoverProvider extends DefaultProviderListener implements Provide
             public String toString() {
                 return "send -> " + envelope;
             }
+
+            @Override
+            public JMSException createTimedOutException() {
+                return new JmsSendTimedOutException("Timed out waiting on " + this, envelope.getMessage());
+            }
         };
 
         serializer.execute(pending);
     }
 
     @Override
-    public void acknowledge(final JmsSessionId sessionId, AsyncResult request) throws IOException, JMSException {
+    public void acknowledge(final JmsSessionId sessionId, final ACK_TYPE ackType, AsyncResult request) throws IOException, JMSException {
         checkClosed();
-        final FailoverRequest pending = new FailoverRequest(request) {
+        final FailoverRequest pending = new FailoverRequest(request, requestTimeout) {
             @Override
             public void doTask() throws Exception {
-                provider.acknowledge(sessionId, this);
+                provider.acknowledge(sessionId, ackType, this);
             }
 
             @Override
@@ -354,7 +398,7 @@ public class FailoverProvider extends DefaultProviderListener implements Provide
     @Override
     public void acknowledge(final JmsInboundMessageDispatch envelope, final ACK_TYPE ackType, AsyncResult request) throws IOException, JMSException {
         checkClosed();
-        final FailoverRequest pending = new FailoverRequest(request) {
+        final FailoverRequest pending = new FailoverRequest(request, requestTimeout) {
             @Override
             public void doTask() throws Exception {
                 provider.acknowledge(envelope, ackType, this);
@@ -376,12 +420,12 @@ public class FailoverProvider extends DefaultProviderListener implements Provide
     }
 
     @Override
-    public void commit(final JmsSessionId sessionId, AsyncResult request) throws IOException, JMSException, UnsupportedOperationException {
+    public void commit(final JmsTransactionInfo transactionInfo, JmsTransactionInfo nextTransactionInfo, AsyncResult request) throws IOException, JMSException, UnsupportedOperationException {
         checkClosed();
-        final FailoverRequest pending = new FailoverRequest(request) {
+        final FailoverRequest pending = new FailoverRequest(request, requestTimeout) {
             @Override
             public void doTask() throws Exception {
-                provider.commit(sessionId, this);
+                provider.commit(transactionInfo, nextTransactionInfo, this);
             }
 
             @Override
@@ -391,7 +435,14 @@ public class FailoverProvider extends DefaultProviderListener implements Provide
 
             @Override
             public String toString() {
-                return "TX commit -> " + sessionId;
+                return "TX commit -> " + transactionInfo.getId();
+            }
+
+            @Override
+            protected Exception createOfflineFailureException(IOException error) {
+                Exception ex = new TransactionRolledBackException("Commit failed, connection offline: " + error.getMessage());
+                ex.initCause(error);
+                return ex;
             }
         };
 
@@ -399,22 +450,22 @@ public class FailoverProvider extends DefaultProviderListener implements Provide
     }
 
     @Override
-    public void rollback(final JmsSessionId sessionId, AsyncResult request) throws IOException, JMSException, UnsupportedOperationException {
+    public void rollback(final JmsTransactionInfo transactionInfo, JmsTransactionInfo nextTransactionInfo, AsyncResult request) throws IOException, JMSException, UnsupportedOperationException {
         checkClosed();
-        final FailoverRequest pending = new FailoverRequest(request) {
+        final FailoverRequest pending = new FailoverRequest(request, requestTimeout) {
             @Override
             public void doTask() throws Exception {
-                provider.rollback(sessionId, this);
+                provider.rollback(transactionInfo, nextTransactionInfo, this);
             }
 
             @Override
-            public boolean failureWhenOffline() {
+            public boolean succeedsWhenOffline() {
                 return true;
             }
 
             @Override
             public String toString() {
-                return "TX rollback -> " + sessionId;
+                return "TX rollback -> " + transactionInfo.getId();
             }
         };
 
@@ -424,7 +475,7 @@ public class FailoverProvider extends DefaultProviderListener implements Provide
     @Override
     public void recover(final JmsSessionId sessionId, final AsyncResult request) throws IOException, UnsupportedOperationException {
         checkClosed();
-        final FailoverRequest pending = new FailoverRequest(request) {
+        final FailoverRequest pending = new FailoverRequest(request, requestTimeout) {
             @Override
             public void doTask() throws Exception {
                 provider.recover(sessionId, this);
@@ -447,7 +498,7 @@ public class FailoverProvider extends DefaultProviderListener implements Provide
     @Override
     public void unsubscribe(final String subscription, AsyncResult request) throws IOException, JMSException, UnsupportedOperationException {
         checkClosed();
-        final FailoverRequest pending = new FailoverRequest(request) {
+        final FailoverRequest pending = new FailoverRequest(request, requestTimeout) {
             @Override
             public void doTask() throws Exception {
                 provider.unsubscribe(subscription, this);
@@ -498,38 +549,51 @@ public class FailoverProvider extends DefaultProviderListener implements Provide
      *        the error that triggered the failure of the provider.
      */
     private void handleProviderFailure(final IOException cause) {
-        LOG.debug("handling Provider failure: {}", cause.getMessage());
-        LOG.trace("stack", cause);
+        if (provider != null) {
+            LOG.debug("handling Provider failure: {}", cause.getMessage());
+            LOG.trace("stack", cause);
 
-        this.provider.setProviderListener(closedListener);
-        URI failedURI = this.provider.getRemoteURI();
-        try {
-            this.provider.close();
-        } catch (Throwable error) {
-            LOG.trace("Caught exception while closing failed provider: {}", error.getMessage());
-        }
-        this.provider = null;
+            provider.setProviderListener(closedListener);
+            URI failedURI = this.provider.getRemoteURI();
+            try {
+                provider.close();
+            } catch (Throwable error) {
+                LOG.trace("Caught exception while closing failed provider: {}", error.getMessage());
+            }
+            provider = null;
 
-        if (reconnectAllowed()) {
+            if (reconnectAllowed(cause)) {
 
-            if (cause instanceof ProviderRedirectedException) {
-                ProviderRedirectedException redirect = (ProviderRedirectedException) cause;
-                try {
-                    uris.addFirst(new URI(failedURI.getScheme() + "://" + redirect.getNetworkHost() + ":" + redirect.getPort()));
-                } catch (URISyntaxException ex) {
-                    LOG.warn("Could not construct redirection URI from remote provided information");
+                if (cause instanceof ProviderRedirectedException) {
+                    ProviderRedirectedException redirect = (ProviderRedirectedException) cause;
+                    try {
+                        uris.addFirst(redirect.getRedirectionURI());
+                    } catch (Exception error) {
+                        LOG.warn("Could not construct redirection URI from remote provided information");
+                    }
                 }
-            }
 
-            ProviderListener listener = this.listener;
-            if (listener != null) {
-                listener.onConnectionInterrupted(failedURI);
-            }
-            triggerReconnectionAttempt();
-        } else {
-            ProviderListener listener = this.listener;
-            if (listener != null) {
-                listener.onConnectionFailure(cause);
+                ProviderListener listener = this.listener;
+                if (listener != null) {
+                    listener.onConnectionInterrupted(failedURI);
+                }
+
+                // Start watching for request timeouts while we are offline, unless we already are.
+                if (requestTimeoutTask == null) {
+                    long sweeperInterval = getRequestSweeperInterval();
+                    if (sweeperInterval > 0) {
+                        LOG.trace("Request timeout monitoring enabled: interval = {}ms", sweeperInterval);
+                        requestTimeoutTask = serializer.scheduleWithFixedDelay(
+                            new FailoverRequestSweeper(), sweeperInterval, sweeperInterval, TimeUnit.MILLISECONDS);
+                    }
+                }
+
+                triggerReconnectionAttempt();
+            } else {
+                ProviderListener listener = this.listener;
+                if (listener != null) {
+                    listener.onConnectionFailure(cause);
+                }
             }
         }
     }
@@ -577,6 +641,16 @@ public class FailoverProvider extends DefaultProviderListener implements Provide
                     reconnectAttempts = 0;
                     connectedURI = provider.getRemoteURI();
                     uris.connected();
+
+                    // Cancel timeout processing since we are connected again.  We waited until
+                    // now for the case where we are continually getting bounced from otherwise
+                    // live servers, we want the timeout to remain scheduled in that case so that
+                    // it doesn't keep getting rescheduled and never actually time anything out.
+                    if (requestTimeoutTask != null) {
+                        requestTimeoutTask.cancel(false);
+                        requestTimeoutTask = null;
+                    }
+
                 } catch (Throwable error) {
                     handleProviderFailure(IOExceptionSupport.create(error));
                 }
@@ -603,6 +677,8 @@ public class FailoverProvider extends DefaultProviderListener implements Provide
         }
 
         connectionHub.execute(new Runnable() {
+            boolean delayed = false;
+
             @Override
             public void run() {
                 if (provider != null || closingConnection.get() || closed.get() || failed.get()) {
@@ -614,7 +690,11 @@ public class FailoverProvider extends DefaultProviderListener implements Provide
                     return;
                 }
 
-                if (initialReconnectDelay > 0 && reconnectAttempts == 0) {
+                boolean first = firstAttempt;
+                firstAttempt = false;
+
+                if (!delayed && !first && initialReconnectDelay > 0 && reconnectAttempts == 0) {
+                    delayed = true;
                     LOG.trace("Delayed initial reconnect attempt will be in {} milliseconds", initialReconnectDelay);
                     connectionHub.schedule(this, initialReconnectDelay, TimeUnit.MILLISECONDS);
                     return;
@@ -628,14 +708,16 @@ public class FailoverProvider extends DefaultProviderListener implements Provide
                     try {
                         LOG.debug("Connection attempt:[{}] to: {} in-progress", reconnectAttempts, target);
                         provider = ProviderFactory.create(target);
-                        provider.connect();
+                        provider.connect(connectionInfo);
                         initializeNewConnection(provider);
                         return;
                     } catch (Throwable e) {
                         LOG.info("Connection attempt:[{}] to: {} failed", reconnectAttempts, target);
                         failure = e;
                         try {
-                            provider.close();
+                            if (provider != null) {
+                                provider.close();
+                            }
                         } catch (Throwable ex) {}
                     }
                 } else {
@@ -669,13 +751,23 @@ public class FailoverProvider extends DefaultProviderListener implements Provide
         });
     }
 
-    private boolean reconnectAllowed() {
+    private boolean reconnectAllowed(IOException cause) {
+        // If a connection attempts fail due to Security errors than
+        // we abort reconnection as there is a configuration issue and
+        // we want to avoid a spinning reconnect cycle that can never
+        // complete.
+        if (cause.getCause() instanceof JMSSecurityException) {
+            return false;
+        }
+
         return reconnectAttemptLimit() != 0;
     }
 
     private int reconnectAttemptLimit() {
         int maxReconnectValue = this.maxReconnectAttempts;
-        if (firstConnection && this.startupMaxReconnectAttempts != UNLIMITED) {
+        if (firstConnection && this.startupMaxReconnectAttempts != UNDEFINED) {
+            // If this is the first connection and a specific startup retry limit
+            // is configured then use it, otherwise use the main reconnect limit
             maxReconnectValue = this.startupMaxReconnectAttempts;
         }
         return maxReconnectValue;
@@ -711,14 +803,26 @@ public class FailoverProvider extends DefaultProviderListener implements Provide
         if (closingConnection.get() || closed.get() || failed.get()) {
             return;
         }
-        serializer.execute(new Runnable() {
-            @Override
-            public void run() {
-                if (!closingConnection.get() && !closed.get() && !failed.get()) {
-                    listener.onInboundMessage(envelope);
-                }
-            }
-        });
+
+        listener.onInboundMessage(envelope);
+    }
+
+    @Override
+    public void onCompletedMessageSend(final JmsOutboundMessageDispatch envelope) {
+        if (closingConnection.get() || closed.get() || failed.get()) {
+            return;
+        }
+
+        listener.onCompletedMessageSend(envelope);
+    }
+
+    @Override
+    public void onFailedMessageSend(final JmsOutboundMessageDispatch envelope, Throwable cause) {
+        if (closingConnection.get() || closed.get() || failed.get()) {
+            return;
+        }
+
+        listener.onFailedMessageSend(envelope, cause);
     }
 
     @Override
@@ -737,6 +841,73 @@ public class FailoverProvider extends DefaultProviderListener implements Provide
         });
     }
 
+    @Override
+    public void onProviderException(final Exception ex) {
+        if (closingConnection.get() || closed.get() || failed.get()) {
+            return;
+        }
+        serializer.execute(new Runnable() {
+            @Override
+            public void run() {
+                if (!closingConnection.get() && !closed.get() && !failed.get()) {
+                    LOG.debug("Failover: the provider reports an async error: {}", ex.getMessage());
+                    listener.onProviderException(ex);
+                }
+            }
+        });
+    }
+
+    @Override
+    public void onRemoteDiscovery(final List<URI> discovered) {
+        if (closingConnection.get() || closed.get() || failed.get()) {
+            return;
+        }
+
+        if (discovered == null || discovered.isEmpty()) {
+            return;
+        }
+
+        serializer.execute(new Runnable() {
+            @Override
+            public void run() {
+                if (!closingConnection.get() && !closed.get() && !failed.get()) {
+
+                    List<URI> newRemotes = new ArrayList<URI>(discovered);
+                    switch (amqpOpenServerListAction) {
+                        case ADD:
+                            try {
+                                uris.addAll(discovered);
+                            } catch (Throwable err) {
+                                LOG.warn("Error while attempting to add discovered URIs: {}", discovered);
+                            }
+                            break;
+                        case REPLACE:
+                            // The current server is assumed not to be in the list of updated remote
+                            // as it is meant for the failover nodes. The pool will de-dup if it is.
+                            newRemotes.add(0, connectedURI);
+                            try {
+                                uris.replaceAll(newRemotes);
+                            } catch (Throwable err) {
+                                LOG.warn("Error while attempting to add discovered URIs: {}", discovered);
+                            }
+                            break;
+                        case IGNORE:
+                            // Do Nothing
+                            break;
+                        default:
+                            // Shouldnt get here, but do nothing if we do.
+                            break;
+                    }
+
+                    // Inform any listener that we've made a new discovery.
+                    if (listener != null) {
+                        listener.onRemoteDiscovery(discovered);
+                    }
+                }
+            }
+        });
+    }
+
     //--------------- URI update and rebalance methods -----------------------//
 
     public void add(final URI uri) {
@@ -867,10 +1038,6 @@ public class FailoverProvider extends DefaultProviderListener implements Provide
         this.useReconnectBackOff = useReconnectBackOff;
     }
 
-    public long getConnectTimeout() {
-        return this.connectTimeout;
-    }
-
     public long getCloseTimeout() {
         return this.closeTimeout;
     }
@@ -883,6 +1050,14 @@ public class FailoverProvider extends DefaultProviderListener implements Provide
         return this.requestTimeout;
     }
 
+    public String getAmqpOpenServerListAction() {
+        return amqpOpenServerListAction.toString();
+    }
+
+    public void setAmqpOpenServerListAction(String amqpOpenServerListAction) {
+        this.amqpOpenServerListAction = FailoverServerListAction.valueOf(amqpOpenServerListAction.toUpperCase(Locale.ENGLISH));
+    }
+
     public Map<String, String> getNestedOptions() {
         return uris.getNestedOptions();
     }
@@ -893,19 +1068,58 @@ public class FailoverProvider extends DefaultProviderListener implements Provide
                (connectedURI == null ? "unconnected" : connectedURI.toString());
     }
 
-    //--------------- FailoverProvider Asynchronous Request --------------------//
+    //--------------- FailoverProvider Request Timeout Support ---------------//
+
+    protected final long getRequestSweeperInterval() {
+        long[] timeouts = new long[] { requestTimeout, sendTimeout };
+
+        Arrays.sort(timeouts);
+
+        for (long timeout : timeouts) {
+            if (timeout != JmsConnectionInfo.INFINITE) {
+                return Math.max(Math.max(1, timeout) / 3, MINIMUM_TIMEOUT);
+            }
+        }
+
+        return DISABLED;
+    }
+
+    protected final class FailoverRequestSweeper implements Runnable {
+
+        @Override
+        public void run() {
+            List<FailoverRequest> copied = new ArrayList<FailoverRequest>(requests.values());
+            for (FailoverRequest request : copied) {
+                if (request.isExpired()) {
+                    LOG.trace("Task {} has timed out, sending failure notice.", request);
+                    request.onFailure(request.createTimedOutException());
+                }
+            }
+        }
+    }
+
+    //--------------- FailoverProvider Asynchronous Request ------------------//
 
     /**
      * For all requests that are dispatched from the FailoverProvider to a connected
      * Provider instance an instance of FailoverRequest is used to handle errors that
      * occur during processing of that request and trigger a reconnect.
      */
-    protected abstract class FailoverRequest extends ProviderFuture implements Runnable {
+    protected abstract class FailoverRequest extends WrappedAsyncResult implements Runnable {
 
-        private final long id = requestId.incrementAndGet();
+        protected final long id = requestId.incrementAndGet();
+
+        private final long requestStarted = System.nanoTime();
+        private final long requestTimeout;
 
         public FailoverRequest(AsyncResult watcher) {
+            this(watcher, JmsConnectionInfo.INFINITE);
+        }
+
+        public FailoverRequest(AsyncResult watcher, long requestTimeout) {
             super(watcher);
+            this.requestTimeout = requestTimeout;
+            LOG.trace("Created Failover Task: {} ({})", this, id);
         }
 
         @Override
@@ -915,7 +1129,7 @@ public class FailoverProvider extends DefaultProviderListener implements Provide
                 whenOffline(new IOException("Connection failed."));
             } else {
                 try {
-                    LOG.debug("Executing Failover Task: {}", this);
+                    LOG.debug("Executing Failover Task: {} ({})", this, id);
                     doTask();
                 } catch (UnsupportedOperationException e) {
                     requests.remove(id);
@@ -932,16 +1146,18 @@ public class FailoverProvider extends DefaultProviderListener implements Provide
         }
 
         @Override
-        public void onFailure(final Throwable result) {
-            if (closingConnection.get() || closed.get() || failed.get()) {
+        public void onFailure(final Throwable error) {
+            if (error instanceof JMSException || closingConnection.get() || closed.get() || failed.get()) {
                 requests.remove(id);
-                super.onFailure(result);
+                super.onFailure(error);
             } else {
-                LOG.debug("Request received error: {}", result.getMessage());
+                LOG.debug("Request received error: {}", error.getMessage());
                 serializer.execute(new Runnable() {
                     @Override
                     public void run() {
-                        handleProviderFailure(IOExceptionSupport.create(result));
+                        IOException ioError = IOExceptionSupport.create(error);
+                        whenOffline(ioError);
+                        handleProviderFailure(ioError);
                     }
                 });
             }
@@ -978,14 +1194,33 @@ public class FailoverProvider extends DefaultProviderListener implements Provide
             return false;
         }
 
+        /**
+         * @return true if the request has a configured expiration time.
+         */
+        public boolean isExpired() {
+            if (requestTimeout != JmsConnectionInfo.INFINITE) {
+                return (System.nanoTime() - requestStarted) > TimeUnit.MILLISECONDS.toNanos(requestTimeout);
+            } else {
+                return false;
+            }
+        }
+
+        protected JMSException createTimedOutException() {
+            return new JmsOperationTimedOutException("Timed out waiting on " +  this);
+        }
+
+        protected Exception createOfflineFailureException(IOException error) {
+            return IOExceptionSupport.create(error);
+        }
+
         private void whenOffline(IOException error) {
             if (failureWhenOffline()) {
                 requests.remove(id);
-                getWrappedRequest().onFailure(IOExceptionSupport.create(error));
+                getWrappedRequest().onFailure(createOfflineFailureException(error));
             } else if (succeedsWhenOffline()) {
                 onSuccess();
             } else {
-                LOG.trace("Task {} held until connection recovered:", this);
+                LOG.trace("Failover task held until connection recovered: {} ({})", this, id);
             }
         }
     }
@@ -1001,9 +1236,6 @@ public class FailoverProvider extends DefaultProviderListener implements Provide
      */
     protected abstract class CreateConnectionRequest extends FailoverRequest {
 
-        /**
-         * @param watcher
-         */
         public CreateConnectionRequest(AsyncResult watcher) {
             super(watcher);
         }
@@ -1027,8 +1259,28 @@ public class FailoverProvider extends DefaultProviderListener implements Provide
             });
         }
 
+        @Override
+        public void onFailure(final Throwable result) {
+            if (closingConnection.get() || closed.get() || failed.get()) {
+                requests.remove(id);
+                super.onFailure(result);
+            } else {
+                LOG.debug("Request received error: {}", result.getMessage());
+                serializer.execute(new Runnable() {
+                    @Override
+                    public void run() {
+                        handleProviderFailure(IOExceptionSupport.create(result));
+                    }
+                });
+            }
+        }
+
         public void signalConnected() {
             super.onSuccess();
         }
     }
+
+    private static enum FailoverServerListAction {
+        ADD, REPLACE, IGNORE
+    };
 }
