@@ -1,4 +1,4 @@
-/**
+/*
  * Licensed to the Apache Software Foundation (ASF) under one or more
  * contributor license agreements.  See the NOTICE file distributed with
  * this work for additional information regarding copyright ownership.
@@ -19,11 +19,18 @@ package org.apache.qpid.jms.meta;
 import static org.junit.Assert.assertEquals;
 import static org.junit.Assert.assertFalse;
 import static org.junit.Assert.assertNotNull;
+import static org.junit.Assert.assertNotSame;
+import static org.junit.Assert.assertNull;
 import static org.junit.Assert.assertSame;
 import static org.junit.Assert.assertTrue;
 
+import java.nio.charset.StandardCharsets;
 import java.util.concurrent.atomic.AtomicBoolean;
 
+import org.apache.qpid.jms.policy.JmsDefaultMessageIDPolicy;
+import org.apache.qpid.jms.policy.JmsDefaultPrefetchPolicy;
+import org.apache.qpid.jms.policy.JmsDefaultPresettlePolicy;
+import org.apache.qpid.jms.policy.JmsDefaultRedeliveryPolicy;
 import org.apache.qpid.jms.util.IdGenerator;
 import org.junit.Before;
 import org.junit.Test;
@@ -52,7 +59,7 @@ public class JmsConnectionInfoTest {
     @Test
     public void testCreate() {
         JmsConnectionInfo info = new JmsConnectionInfo(firstId);
-        assertSame(firstId, info.getConnectionId());
+        assertSame(firstId, info.getId());
         assertNotNull(info.toString());
     }
 
@@ -60,24 +67,32 @@ public class JmsConnectionInfoTest {
     public void testCopy() {
         JmsConnectionInfo info = new JmsConnectionInfo(firstId);
 
-        info.setAlwaysSyncSend(true);
-        info.setClientId("test");
+        info.setForceSyncSend(true);
+        info.setClientId("test", true);
         info.setCloseTimeout(100);
         info.setConnectTimeout(200);
-        info.setForceAsyncSends(true);
+        info.setForceAsyncSend(true);
         info.setPassword("pass");
         info.setQueuePrefix("queue");
         info.setRequestTimeout(50);
         info.setSendTimeout(150);
         info.setTopicPrefix("topic");
         info.setUsername("user");
-        boolean validatePropertyNames = ! info.isValidatePropertyNames();
+        boolean validatePropertyNames = !info.isValidatePropertyNames();
         info.setValidatePropertyNames(validatePropertyNames);
+        boolean awaitClientID = !info.isAwaitClientID();
+        info.setAwaitClientID(awaitClientID);
+        info.setMessageIDPolicy(new JmsDefaultMessageIDPolicy());
+        info.setPrefetchPolicy(new JmsDefaultPrefetchPolicy());
+        info.setPresettlePolicy(new JmsDefaultPresettlePolicy());
+        info.setRedeliveryPolicy(new JmsDefaultRedeliveryPolicy());
 
         JmsConnectionInfo copy = info.copy();
 
-        assertEquals(true, copy.isAlwaysSyncSend());
+        assertEquals(true, copy.isForceSyncSend());
         assertEquals("test", copy.getClientId());
+        assertEquals(true, copy.isExplicitClientID());
+        assertEquals(awaitClientID, copy.isAwaitClientID());
         assertEquals(100, copy.getCloseTimeout());
         assertEquals(200, copy.getConnectTimeout());
         assertEquals(true, copy.isForceAsyncSend());
@@ -89,17 +104,20 @@ public class JmsConnectionInfoTest {
         assertEquals("user", copy.getUsername());
         assertEquals(validatePropertyNames, copy.isValidatePropertyNames());
 
+        assertNotSame(info.getPrefetchPolicy(), copy.getPrefetchPolicy());
+        assertNotSame(info.getPresettlePolicy(), copy.getPresettlePolicy());
+        assertNotSame(info.getRedeliveryPolicy(), copy.getRedeliveryPolicy());
+        assertNotSame(info.getMessageIDPolicy(), copy.getMessageIDPolicy());
+
         assertEquals(info, copy);
-}
+    }
 
     @Test
     public void testCompareTo() {
         JmsConnectionInfo first = new JmsConnectionInfo(firstId);
         JmsConnectionInfo second = new JmsConnectionInfo(secondId);
 
-        assertEquals(-1, first.compareTo(second));
-        assertEquals(0, first.compareTo(first));
-        assertEquals(1, second.compareTo(first));
+        assertFalse(first.compareTo(second) == 0);
     }
 
     @Test
@@ -145,4 +163,42 @@ public class JmsConnectionInfoTest {
 
         assertTrue(visited.get());
     }
+
+    @Test
+    public void testIsExplicitClientId() {
+        final JmsConnectionInfo info = new JmsConnectionInfo(firstId);
+        assertFalse(info.isExplicitClientID());
+        info.setClientId("something", true);
+        assertTrue(info.isExplicitClientID());
+    }
+
+    @Test
+    public void testGetEncodedUsername() {
+        final JmsConnectionInfo info = new JmsConnectionInfo(firstId);
+        info.setUsername("user");
+        byte[] result = info.getEncodedUsername();
+        assertEquals("user", new String(result, StandardCharsets.UTF_8));
+    }
+
+    @Test
+    public void testGetEncodedUsernameValueIsCached() {
+        final JmsConnectionInfo info = new JmsConnectionInfo(firstId);
+        info.setUsername("user");
+        byte[] result1 = info.getEncodedUsername();
+        byte[] result2 = info.getEncodedUsername();
+        assertSame(result1, result2);
+    }
+
+    @Test
+    public void testGetEncodedUsernameWithNoUsername() {
+        final JmsConnectionInfo info = new JmsConnectionInfo(firstId);
+        assertNull(info.getEncodedUsername());
+    }
+
+    @Test
+    public void testGetEncodedUsernameWithEmptyUsername() {
+        final JmsConnectionInfo info = new JmsConnectionInfo(firstId);
+        info.setUsername("");
+        assertNotNull(info.getEncodedUsername());
+    }
 }
