@@ -1,4 +1,4 @@
-/**
+/*
  * Licensed to the Apache Software Foundation (ASF) under one or more
  * contributor license agreements.  See the NOTICE file distributed with
  * this work for additional information regarding copyright ownership.
@@ -63,6 +63,7 @@ public class FailoverProviderTest extends FailoverProviderTestSupport {
 
     private List<URI> uris;
     private FailoverProvider provider;
+    private JmsConnectionInfo connection;
 
     @Override
     @Before
@@ -74,6 +75,8 @@ public class FailoverProviderTest extends FailoverProviderTestSupport {
         uris.add(new URI("mock://192.168.2.3:5672"));
         uris.add(new URI("mock://192.168.2.4:5672"));
 
+        connection = createConnectionInfo();
+
         super.setUp();
     }
 
@@ -130,7 +133,7 @@ public class FailoverProviderTest extends FailoverProviderTestSupport {
         provider = new FailoverProvider(uris, Collections.<String, String>emptyMap());
 
         assertNull(provider.getRemoteURI());
-        provider.connect();
+        provider.connect(connection);
         assertTrue("Should have a remote URI after connect", Wait.waitFor(new Wait.Condition() {
 
             @Override
@@ -140,40 +143,12 @@ public class FailoverProviderTest extends FailoverProviderTestSupport {
         }, TimeUnit.SECONDS.toMillis(20), 10));
     }
 
-    @Test(timeout = 15000)
-    public void testGetLocalPrincipal() throws Exception {
-        String principalName = "foo";
-
-        FailoverProviderFactory factory = new FailoverProviderFactory();
-        provider = (FailoverProvider) factory.createProvider(new URI("failover:(mock://192.168.2.1:5672?mock.localPrincipal=" + principalName + ")"));
-        provider.setProviderListener(new DefaultProviderListener());
-
-        provider.connect();
-
-        ProviderFuture request = new ProviderFuture();
-        provider.create(createConnectionInfo(), request);
-
-        request.sync(10, TimeUnit.SECONDS);
-        assertTrue(request.isComplete());
-
-        assertNotNull(provider.getLocalPrincipal());
-        assertEquals(principalName, provider.getLocalPrincipal().getName());
-    }
-
-    @Test(timeout = 15000)
-    public void testGetLocalPrincipalNull() throws Exception {
-        FailoverProviderFactory factory = new FailoverProviderFactory();
-        provider = (FailoverProvider) factory.createProvider(new URI("failover:(mock://192.168.2.1:5672)"));
-
-        assertNull(provider.getLocalPrincipal());
-    }
-
     @Test(timeout = 30000)
     public void testToString() throws Exception {
         provider = new FailoverProvider(uris, Collections.<String, String>emptyMap());
 
         assertNotNull(provider.toString());
-        provider.connect();
+        provider.connect(connection);
         assertTrue("Should have a mock scheme after connect", Wait.waitFor(new Wait.Condition() {
 
             @Override
@@ -200,7 +175,7 @@ public class FailoverProviderTest extends FailoverProviderTestSupport {
             }
         });
 
-        provider.connect();
+        provider.connect(connection);
 
         ProviderFuture request = new ProviderFuture();
         provider.create(createConnectionInfo(), request);
@@ -220,7 +195,7 @@ public class FailoverProviderTest extends FailoverProviderTestSupport {
         provider = new FailoverProvider(uris, Collections.<String, String>emptyMap());
         assertEquals(FailoverUriPool.DEFAULT_RANDOMIZE_ENABLED, provider.isRandomize());
         assertNull(provider.getRemoteURI());
-        provider.connect();
+        provider.connect(connection);
 
         try {
             provider.start();
@@ -238,14 +213,16 @@ public class FailoverProviderTest extends FailoverProviderTestSupport {
             "?failover.startupMaxReconnectAttempts=5" +
             "&failover.useReconnectBackOff=false");
 
-        Connection connection = factory.createConnection();
-
+        Connection connection = null;
         try {
+            connection = factory.createConnection();
             connection.start();
             fail("Should have stopped after five retries.");
         } catch (JMSException ex) {
         } finally {
-            connection.close();
+            if (connection != null) {
+                connection.close();
+            }
         }
 
         assertEquals(5, mockPeer.getContextStats().getProvidersCreated());
@@ -260,14 +237,16 @@ public class FailoverProviderTest extends FailoverProviderTestSupport {
             "?failover.maxReconnectAttempts=5" +
             "&failover.useReconnectBackOff=false");
 
-        Connection connection = factory.createConnection();
-
+        Connection connection = null;
         try {
+            connection = factory.createConnection();
             connection.start();
             fail("Should have stopped after five retries.");
         } catch (JMSException ex) {
         } finally {
-            connection.close();
+            if (connection != null) {
+                connection.close();
+            }
         }
 
         assertEquals(5, mockPeer.getContextStats().getProvidersCreated());
@@ -284,14 +263,16 @@ public class FailoverProviderTest extends FailoverProviderTestSupport {
             "&failover.reconnectDelay=10" +
             "&failover.useReconnectBackOff=true");
 
-        Connection connection = factory.createConnection();
-
+        Connection connection = null;
         try {
+            connection = factory.createConnection();
             connection.start();
             fail("Should have stopped after five retries.");
         } catch (JMSException ex) {
         } finally {
-            connection.close();
+            if (connection != null) {
+                connection.close();
+            }
         }
 
         assertEquals(5, mockPeer.getContextStats().getProvidersCreated());
@@ -420,7 +401,7 @@ public class FailoverProviderTest extends FailoverProviderTestSupport {
             }
         });
 
-        provider.connect();
+        provider.connect(connection);
         provider.start();
 
         JmsConnectionInfo connectionInfo = createConnectionInfo();
@@ -434,9 +415,35 @@ public class FailoverProviderTest extends FailoverProviderTestSupport {
         provider.create(connectionInfo, request);
         request.sync();
 
-        assertEquals(CONNECT_TIMEOUT, provider.getConnectTimeout());
         assertEquals(CLOSE_TIMEOUT, provider.getCloseTimeout());
         assertEquals(SEND_TIMEOUT, provider.getSendTimeout());
         assertEquals(REQUEST_TIMEOUT, provider.getRequestTimeout());
     }
+
+    @Test(timeout = 30000)
+    public void testAmqpOpenServerListActionDefault() {
+        provider = new FailoverProvider(uris);
+        assertEquals("REPLACE", provider.getAmqpOpenServerListAction());
+    }
+
+    @Test(timeout = 30000)
+    public void testSetGetAmqpOpenServerListAction() {
+        provider = new FailoverProvider(uris);
+        String action = "ADD";
+        assertFalse(action.equals(provider.getAmqpOpenServerListAction()));
+
+        provider.setAmqpOpenServerListAction(action);
+        assertEquals(action, provider.getAmqpOpenServerListAction());
+    }
+
+    @Test(timeout = 30000)
+    public void testSetInvalidAmqpOpenServerListActionThrowsIAE() {
+        provider = new FailoverProvider(uris);
+        try {
+            provider.setAmqpOpenServerListAction("invalid");
+            fail("no exception was thrown");
+        } catch (IllegalArgumentException iae) {
+            // Expected
+        }
+    }
 }
