@@ -29,7 +29,9 @@ import javax.net.ssl.SSLContext;
 import javax.net.ssl.SSLServerSocket;
 import javax.net.ssl.SSLServerSocketFactory;
 
+import org.apache.qpid.jms.test.testpeer.basictypes.AmqpError;
 import org.apache.qpid.proton.amqp.Binary;
+import org.apache.qpid.proton.engine.impl.AmqpHeader;
 import org.slf4j.Logger;
 import org.slf4j.LoggerFactory;
 
@@ -46,9 +48,11 @@ class TestAmqpPeerRunner implements Runnable
     private OutputStream _networkOutputStream;
 
     private final Object _inputHandlingLock = new Object();
+    private final TestAmqpPeer _peer;
     private final TestFrameParser _testFrameParser;
     private volatile boolean _suppressReadExceptionOnClose;
     private volatile boolean _exitReadLoopEarly;
+    private volatile boolean _sendSaslHeaderPreEmptively;
 
     private volatile Throwable _throwable;
 
@@ -76,6 +80,7 @@ class TestAmqpPeerRunner implements Runnable
         }
 
         _testFrameParser = new TestFrameParser(peer);
+        _peer = peer;
     }
 
     @Override
@@ -90,8 +95,15 @@ class TestAmqpPeerRunner implements Runnable
         )
         {
             _clientSocket = clientSocket;
+            _clientSocket.setTcpNoDelay(true);
             _networkOutputStream = networkOutputStream;
 
+            if (_sendSaslHeaderPreEmptively) {
+                byte[] bytes = AmqpHeader.SASL_HEADER;
+                LOGGER.debug("Sending header pre-emptively: {}", new Binary(bytes));
+                _networkOutputStream.write(bytes);
+            }
+
             int bytesRead;
             byte[] networkInputBytes = new byte[1024];
 
@@ -107,7 +119,17 @@ class TestAmqpPeerRunner implements Runnable
 
                     LOGGER.debug("Read: {}", new Binary(networkInputBytes, 0, bytesRead));
 
-                    _testFrameParser.input(networkInputByteBuffer);
+                    try {
+                        _testFrameParser.input(networkInputByteBuffer);
+                    } catch (Exception ex) {
+                        try {
+                            _peer.sendConnectionCloseImmediately(AmqpError.INTERNAL_ERROR, "Problem in peer: " + ex);
+                        } catch (Exception ex2) {
+                            LOGGER.debug("Exception while sending close frame during handling of previous exception", ex2);
+                        }
+
+                        throw ex;
+                    }
                 }
 
                 if(_exitReadLoopEarly)
@@ -185,7 +207,7 @@ class TestAmqpPeerRunner implements Runnable
 
     public void sendBytes(byte[] bytes)
     {
-        LOGGER.debug("Sending: {}", new Binary(bytes));
+        LOGGER.debug("Sending: {} ({} bytes)", new Binary(bytes), bytes.length);
         try
         {
             _networkOutputStream.write(bytes);
@@ -229,4 +251,16 @@ class TestAmqpPeerRunner implements Runnable
     public void exitReadLoopEarly() {
         _exitReadLoopEarly = true;
     }
+
+    public void setSendSaslHeaderPreEmptively(boolean sendSaslHeaderPreEmptively) {
+        _sendSaslHeaderPreEmptively = sendSaslHeaderPreEmptively;
+    }
+
+    public boolean isSendSaslHeaderPreEmptively() {
+        return _sendSaslHeaderPreEmptively;
+    }
+
+    public boolean isSSL() {
+        return _serverSocket instanceof SSLServerSocket;
+    }
 }
