@@ -1,4 +1,4 @@
-/**
+/*
  * Licensed to the Apache Software Foundation (ASF) under one or more
  * contributor license agreements.  See the NOTICE file distributed with
  * this work for additional information regarding copyright ownership.
@@ -22,10 +22,10 @@ import static org.junit.Assert.assertNotNull;
 import static org.junit.Assert.assertNull;
 import static org.junit.Assert.assertTrue;
 import static org.junit.Assert.fail;
-import io.netty.buffer.ByteBuf;
-import io.netty.buffer.Unpooled;
+import static org.junit.Assume.assumeTrue;
 
 import java.io.IOException;
+import java.lang.reflect.Field;
 import java.net.URI;
 import java.util.ArrayList;
 import java.util.List;
@@ -40,6 +40,11 @@ import org.junit.Test;
 import org.slf4j.Logger;
 import org.slf4j.LoggerFactory;
 
+import io.netty.buffer.ByteBuf;
+import io.netty.buffer.Unpooled;
+import io.netty.channel.epoll.Epoll;
+import io.netty.channel.epoll.EpollEventLoopGroup;
+
 /**
  * Test basic functionality of the Netty based TCP transport.
  */
@@ -71,16 +76,19 @@ public class NettyTcpTransportTest extends QpidJmsTestCase {
     }
 
     @Test(timeout = 60 * 1000)
-    public void testCreateWithNullOptionsUsesDefaults() throws Exception {
+    public void testCreateWithNullOptionsThrowsIAE() throws Exception {
         URI serverLocation = new URI("tcp://localhost:5762");
 
-        Transport transport = createTransport(serverLocation, testListener, null);
-        assertEquals(TransportOptions.INSTANCE, transport.getTransportOptions());
+        try {
+            createTransport(serverLocation, testListener, null);
+            fail("Should have thrown IllegalArgumentException");
+        } catch (IllegalArgumentException iae) {
+        }
     }
 
     @Test(timeout = 60 * 1000)
     public void testConnectWithoutRunningServer() throws Exception {
-        try (NettyEchoServer server = new NettyEchoServer(createServerOptions())) {
+        try (NettyEchoServer server = createEchoServer(createServerOptions())) {
             server.start();
 
             int port = server.getServerPort();
@@ -90,7 +98,7 @@ public class NettyTcpTransportTest extends QpidJmsTestCase {
 
             Transport transport = createTransport(serverLocation, testListener, createClientOptions());
             try {
-                transport.connect();
+                transport.connect(null);
                 fail("Should have failed to connect to the server: " + serverLocation);
             } catch (Exception e) {
                 LOG.info("Failed to connect to: {} as expected.", serverLocation);
@@ -108,7 +116,7 @@ public class NettyTcpTransportTest extends QpidJmsTestCase {
 
     @Test(timeout = 60 * 1000)
     public void testConnectWithoutListenerFails() throws Exception {
-        try (NettyEchoServer server = new NettyEchoServer(createServerOptions())) {
+        try (NettyEchoServer server = createEchoServer(createServerOptions())) {
             server.start();
 
             int port = server.getServerPort();
@@ -116,7 +124,7 @@ public class NettyTcpTransportTest extends QpidJmsTestCase {
 
             Transport transport = createTransport(serverLocation, null, createClientOptions());
             try {
-                transport.connect();
+                transport.connect(null);
                 fail("Should have failed to connect to the server: " + serverLocation);
             } catch (Exception e) {
                 LOG.info("Failed to connect to: {} as expected.", serverLocation);
@@ -130,7 +138,7 @@ public class NettyTcpTransportTest extends QpidJmsTestCase {
 
     @Test(timeout = 60 * 1000)
     public void testConnectAfterListenerSetWorks() throws Exception {
-        try (NettyEchoServer server = new NettyEchoServer(createServerOptions())) {
+        try (NettyEchoServer server = createEchoServer(createServerOptions())) {
             server.start();
 
             int port = server.getServerPort();
@@ -142,10 +150,10 @@ public class NettyTcpTransportTest extends QpidJmsTestCase {
             assertNotNull(transport.getTransportListener());
 
             try {
-                transport.connect();
+                transport.connect(null);
                 LOG.info("Connected to server:{} as expected.", serverLocation);
             } catch (Exception e) {
-                fail("Should not have failed to connect to the server: " + serverLocation);
+                fail("Should not have failed to connect to the server at " + serverLocation + " but got exception: " + e);
             }
 
             assertTrue(transport.isConnected());
@@ -156,7 +164,7 @@ public class NettyTcpTransportTest extends QpidJmsTestCase {
 
     @Test(timeout = 60 * 1000)
     public void testConnectToServer() throws Exception {
-        try (NettyEchoServer server = new NettyEchoServer(createServerOptions())) {
+        try (NettyEchoServer server = createEchoServer(createServerOptions())) {
             server.start();
 
             int port = server.getServerPort();
@@ -164,10 +172,10 @@ public class NettyTcpTransportTest extends QpidJmsTestCase {
 
             Transport transport = createTransport(serverLocation, testListener, createClientOptions());
             try {
-                transport.connect();
+                transport.connect(null);
                 LOG.info("Connected to server:{} as expected.", serverLocation);
             } catch (Exception e) {
-                fail("Should have connected to the server: " + serverLocation);
+                fail("Should have connected to the server at " + serverLocation + " but got exception: " + e);
             }
 
             assertTrue(transport.isConnected());
@@ -188,7 +196,7 @@ public class NettyTcpTransportTest extends QpidJmsTestCase {
     public void testMultipleConnectionsToServer() throws Exception {
         final int CONNECTION_COUNT = 10;
 
-        try (NettyEchoServer server = new NettyEchoServer(createServerOptions())) {
+        try (NettyEchoServer server = createEchoServer(createServerOptions())) {
             server.start();
 
             int port = server.getServerPort();
@@ -199,12 +207,12 @@ public class NettyTcpTransportTest extends QpidJmsTestCase {
             for (int i = 0; i < CONNECTION_COUNT; ++i) {
                 Transport transport = createTransport(serverLocation, testListener, createClientOptions());
                 try {
-                    transport.connect();
+                    transport.connect(null);
                     assertTrue(transport.isConnected());
                     LOG.info("Connected to server:{} as expected.", serverLocation);
                     transports.add(transport);
                 } catch (Exception e) {
-                    fail("Should have connected to the server: " + serverLocation);
+                    fail("Should have connected to the server at " + serverLocation + " but got exception: " + e);
                 }
             }
 
@@ -228,7 +236,7 @@ public class NettyTcpTransportTest extends QpidJmsTestCase {
             sendBuffer.writeByte('A');
         }
 
-        try (NettyEchoServer server = new NettyEchoServer(createServerOptions())) {
+        try (NettyEchoServer server = createEchoServer(createServerOptions())) {
             server.start();
 
             int port = server.getServerPort();
@@ -239,11 +247,11 @@ public class NettyTcpTransportTest extends QpidJmsTestCase {
             for (int i = 0; i < CONNECTION_COUNT; ++i) {
                 Transport transport = createTransport(serverLocation, testListener, createClientOptions());
                 try {
-                    transport.connect();
+                    transport.connect(null);
                     transport.send(sendBuffer.copy());
                     transports.add(transport);
                 } catch (Exception e) {
-                    fail("Should have connected to the server: " + serverLocation);
+                    fail("Should have connected to the server at " + serverLocation + " but got exception: " + e);
                 }
             }
 
@@ -268,7 +276,7 @@ public class NettyTcpTransportTest extends QpidJmsTestCase {
     public void testDetectServerClose() throws Exception {
         Transport transport = null;
 
-        try (NettyEchoServer server = new NettyEchoServer(createServerOptions())) {
+        try (NettyEchoServer server = createEchoServer(createServerOptions())) {
             server.start();
 
             int port = server.getServerPort();
@@ -276,10 +284,10 @@ public class NettyTcpTransportTest extends QpidJmsTestCase {
 
             transport = createTransport(serverLocation, testListener, createClientOptions());
             try {
-                transport.connect();
+                transport.connect(null);
                 LOG.info("Connected to server:{} as expected.", serverLocation);
             } catch (Exception e) {
-                fail("Should have connected to the server: " + serverLocation);
+                fail("Should have connected to the server at " + serverLocation + " but got exception: " + e);
             }
 
             assertTrue(transport.isConnected());
@@ -307,7 +315,7 @@ public class NettyTcpTransportTest extends QpidJmsTestCase {
 
     @Test(timeout = 60 * 1000)
     public void testZeroSizedSentNoErrors() throws Exception {
-        try (NettyEchoServer server = new NettyEchoServer(createServerOptions())) {
+        try (NettyEchoServer server = createEchoServer(createServerOptions())) {
             server.start();
 
             int port = server.getServerPort();
@@ -315,10 +323,10 @@ public class NettyTcpTransportTest extends QpidJmsTestCase {
 
             Transport transport = createTransport(serverLocation, testListener, createClientOptions());
             try {
-                transport.connect();
+                transport.connect(null);
                 LOG.info("Connected to server:{} as expected.", serverLocation);
             } catch (Exception e) {
-                fail("Should have connected to the server: " + serverLocation);
+                fail("Should have connected to the server at " + serverLocation + " but got exception: " + e);
             }
 
             assertTrue(transport.isConnected());
@@ -335,7 +343,7 @@ public class NettyTcpTransportTest extends QpidJmsTestCase {
 
     @Test(timeout = 60 * 1000)
     public void testDataSentIsReceived() throws Exception {
-        try (NettyEchoServer server = new NettyEchoServer(createServerOptions())) {
+        try (NettyEchoServer server = createEchoServer(createServerOptions())) {
             server.start();
 
             int port = server.getServerPort();
@@ -343,10 +351,10 @@ public class NettyTcpTransportTest extends QpidJmsTestCase {
 
             Transport transport = createTransport(serverLocation, testListener, createClientOptions());
             try {
-                transport.connect();
+                transport.connect(null);
                 LOG.info("Connected to server:{} as expected.", serverLocation);
             } catch (Exception e) {
-                fail("Should have connected to the server: " + serverLocation);
+                fail("Should have connected to the server at " + serverLocation + " but got exception: " + e);
             }
 
             assertTrue(transport.isConnected());
@@ -387,7 +395,7 @@ public class NettyTcpTransportTest extends QpidJmsTestCase {
 
     public void doMultipleDataPacketsSentAndReceive(final int byteCount, final int iterations) throws Exception {
 
-        try (NettyEchoServer server = new NettyEchoServer(createServerOptions())) {
+        try (NettyEchoServer server = createEchoServer(createServerOptions())) {
             server.start();
 
             int port = server.getServerPort();
@@ -395,10 +403,10 @@ public class NettyTcpTransportTest extends QpidJmsTestCase {
 
             Transport transport = createTransport(serverLocation, testListener, createClientOptions());
             try {
-                transport.connect();
+                transport.connect(null);
                 LOG.info("Connected to server:{} as expected.", serverLocation);
             } catch (Exception e) {
-                fail("Should have connected to the server: " + serverLocation);
+                fail("Should have connected to the server at " + serverLocation + " but got exception: " + e);
             }
 
             assertTrue(transport.isConnected());
@@ -431,7 +439,7 @@ public class NettyTcpTransportTest extends QpidJmsTestCase {
     public void testSendToClosedTransportFails() throws Exception {
         Transport transport = null;
 
-        try (NettyEchoServer server = new NettyEchoServer(createServerOptions())) {
+        try (NettyEchoServer server = createEchoServer(createServerOptions())) {
             server.start();
 
             int port = server.getServerPort();
@@ -439,10 +447,10 @@ public class NettyTcpTransportTest extends QpidJmsTestCase {
 
             transport = createTransport(serverLocation, testListener, createClientOptions());
             try {
-                transport.connect();
+                transport.connect(null);
                 LOG.info("Connected to server:{} as expected.", serverLocation);
             } catch (Exception e) {
-                fail("Should have connected to the server: " + serverLocation);
+                fail("Should have connected to the server at " + serverLocation + " but got exception: " + e);
             }
 
             assertTrue(transport.isConnected());
@@ -458,6 +466,76 @@ public class NettyTcpTransportTest extends QpidJmsTestCase {
         }
     }
 
+    @Test(timeout = 60 * 1000)
+    public void testConnectToServerWithEpollEnabled() throws Exception {
+        doTestEpollSupport(true);
+    }
+
+    @Test(timeout = 60 * 1000)
+    public void testConnectToServerWithEpollDisabled() throws Exception {
+        doTestEpollSupport(false);
+    }
+
+    private void doTestEpollSupport(boolean useEpoll) throws Exception {
+        assumeTrue(Epoll.isAvailable());
+
+        try (NettyEchoServer server = createEchoServer(createServerOptions())) {
+            server.start();
+
+            int port = server.getServerPort();
+            URI serverLocation = new URI("tcp://localhost:" + port);
+
+            TransportOptions options = createClientOptions();
+            options.setUseEpoll(useEpoll);
+
+            Transport transport = createTransport(serverLocation, testListener, options);
+            try {
+                transport.connect(null);
+                LOG.info("Connected to server:{} as expected.", serverLocation);
+            } catch (Exception e) {
+                fail("Should have connected to the server at " + serverLocation + " but got exception: " + e);
+            }
+
+            assertTrue(transport.isConnected());
+            assertEquals(serverLocation, transport.getRemoteLocation());
+            assertEpoll("Transport should be using Epoll", useEpoll, transport);
+
+            transport.close();
+
+            // Additional close should not fail or cause other problems.
+            transport.close();
+        }
+
+        assertTrue(!transportClosed);  // Normal shutdown does not trigger the event.
+        assertTrue(exceptions.isEmpty());
+        assertTrue(data.isEmpty());
+    }
+
+    private void assertEpoll(String message, boolean expected, Transport transport) throws Exception {
+        Field group = null;
+        Class<?> transportType = transport.getClass();
+
+        while (transportType != null && group == null) {
+            try {
+                group = transportType.getDeclaredField("group");
+            } catch (NoSuchFieldException error) {
+                transportType = transportType.getSuperclass();
+                if (Object.class.equals(transportType)) {
+                    transportType = null;
+                }
+            }
+        }
+
+        assertNotNull("Transport implementation unknown", group);
+
+        group.setAccessible(true);
+        if (expected) {
+            assertTrue(message, group.get(transport) instanceof EpollEventLoopGroup);
+        } else {
+            assertFalse(message, group.get(transport) instanceof EpollEventLoopGroup);
+        }
+    }
+
     protected Transport createTransport(URI serverLocation, TransportListener listener, TransportOptions options) {
         if (listener == null) {
             return new NettyTcpTransport(serverLocation, options);
@@ -467,11 +545,11 @@ public class NettyTcpTransportTest extends QpidJmsTestCase {
     }
 
     protected TransportOptions createClientOptions() {
-        return TransportOptions.INSTANCE.clone();
+        return new TransportOptions();
     }
 
     protected TransportOptions createServerOptions() {
-        return TransportOptions.INSTANCE.clone();
+        return new TransportOptions();
     }
 
     protected void logTransportErrors() {
@@ -482,6 +560,14 @@ public class NettyTcpTransportTest extends QpidJmsTestCase {
         }
     }
 
+    protected NettyEchoServer createEchoServer(TransportOptions options) {
+        return createEchoServer(options, false);
+    }
+
+    protected NettyEchoServer createEchoServer(TransportOptions options, boolean needClientAuth) {
+        return new NettyEchoServer(options, needClientAuth);
+    }
+
     private class NettyTransportListener implements TransportListener {
 
         @Override
